# Generated by Django 4.2.20 on 2025-05-08 10:06

from django.db import migrations, models
import django.db.models.deletion
import django_jsonform.models.fields
import mozilla_django_oidc_db.models


class Migration(migrations.Migration):

    dependencies = [
        ("mozilla_django_oidc_db", "0005_alter_openidconnectconfig_claim_mapping"),
    ]

    operations = [
        migrations.CreateModel(
            name="OIDCProvider",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "identifier",
                    models.SlugField(
                        help_text="Unique identifier of the OIDC provider.",
                        max_length=255,
                        unique=True,
                        verbose_name="identifier",
                    ),
                ),
                (
                    "oidc_op_discovery_endpoint",
                    models.URLField(
                        blank=True,
                        help_text="URL of your provider discovery endpoint ending with a slash (`.well-known/...` will be added automatically). If this is provided, the remaining endpoints can be omitted, as they will be derived from this endpoint.",
                        max_length=1000,
                        verbose_name="Discovery endpoint",
                    ),
                ),
                (
                    "oidc_op_jwks_endpoint",
                    models.URLField(
                        blank=True,
                        help_text="URL of your provider JSON Web Key Set endpoint. Required if `RS256` is used as signing algorithm.",
                        max_length=1000,
                        verbose_name="JSON Web Key Set endpoint",
                    ),
                ),
                (
                    "oidc_op_authorization_endpoint",
                    models.URLField(
                        help_text="URL of your provider authorization endpoint",
                        max_length=1000,
                        verbose_name="Authorization endpoint",
                    ),
                ),
                (
                    "oidc_op_token_endpoint",
                    models.URLField(
                        help_text="URL of your provider token endpoint",
                        max_length=1000,
                        verbose_name="Token endpoint",
                    ),
                ),
                (
                    "oidc_op_user_endpoint",
                    models.URLField(
                        help_text="URL of your provider userinfo endpoint.",
                        max_length=1000,
                        verbose_name="User endpoint",
                    ),
                ),
                (
                    "oidc_op_logout_endpoint",
                    models.URLField(
                        blank=True,
                        help_text="URL of your provider logout endpoint.",
                        max_length=1000,
                        verbose_name="Logout endpoint",
                    ),
                ),
                (
                    "oidc_token_use_basic_auth",
                    models.BooleanField(
                        blank=True,
                        default=False,
                        help_text="If enabled, the client ID and secret are sent in the HTTP Basic auth header when obtaining the access token. Otherwise, they are sent in the request body.",
                        verbose_name="Use Basic auth for token endpoint",
                    ),
                ),
                (
                    "oidc_use_nonce",
                    models.BooleanField(
                        default=True,
                        blank=True,
                        help_text="Controls whether the client uses nonce verification",
                        verbose_name="Use nonce",
                    ),
                ),
                (
                    "oidc_nonce_size",
                    models.PositiveIntegerField(
                        default=32,
                        blank=True,
                        help_text="Sets the length of the random string used for nonce verification",
                        verbose_name="Nonce size",
                    ),
                ),
                (
                    "oidc_state_size",
                    models.PositiveIntegerField(
                        default=32,
                        blank=True,
                        help_text="Sets the length of the random string used for state verification",
                        verbose_name="State size",
                    ),
                ),
            ],
            options={
                "verbose_name": "OIDC Provider",
                "verbose_name_plural": "OIDC Providers",
            },
        ),
        migrations.CreateModel(
            name="OIDCClient",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "identifier",
                    models.SlugField(
                        help_text="Unique identifier for the client.",
                        max_length=255,
                        unique=True,
                        verbose_name="identifier",
                    ),
                ),
                (
                    "enabled",
                    models.BooleanField(
                        default=False,
                        help_text="The client must be enabled before users can authenticate through it.",
                        verbose_name="enabled",
                    ),
                ),
                (
                    "oidc_rp_client_id",
                    models.CharField(
                        help_text="Client ID provided by the OIDC Provider",
                        max_length=1000,
                        verbose_name="Client ID",
                    ),
                ),
                (
                    "oidc_rp_client_secret",
                    models.CharField(
                        help_text="Secret provided by the OIDC Provider",
                        max_length=1000,
                        verbose_name="Secret",
                    ),
                ),
                (
                    "oidc_rp_scopes_list",
                    django_jsonform.models.fields.ArrayField(
                        base_field=models.CharField(
                            max_length=50, verbose_name="Scope"
                        ),
                        blank=True,
                        default=mozilla_django_oidc_db.models.get_default_scopes,
                        help_text="Scopes that are requested during login",
                        size=None,
                        verbose_name="Scopes",
                    ),
                ),
                (
                    "oidc_rp_sign_algo",
                    models.CharField(
                        default="RS256",
                        help_text="Algorithm the Identity Provider uses to sign ID tokens",
                        max_length=50,
                        verbose_name="OpenID sign algorithm",
                    ),
                ),
                (
                    "oidc_rp_idp_sign_key",
                    models.CharField(
                        blank=True,
                        help_text="Key the Identity Provider uses to sign ID tokens in the case of an RSA sign algorithm. Should be the signing key in PEM or DER format.",
                        max_length=1000,
                        verbose_name="Sign key",
                    ),
                ),
                (
                    "oidc_keycloak_idp_hint",
                    models.CharField(
                        blank=True,
                        help_text="Specific for Keycloak: parameter that indicates which identity provider should be used (therefore skipping the Keycloak login screen).",
                        max_length=1000,
                        verbose_name="Keycloak Identity Provider hint",
                    ),
                ),
                (
                    "userinfo_claims_source",
                    models.CharField(
                        choices=[
                            ("userinfo_endpoint", "Userinfo endpoint"),
                            ("id_token", "ID token"),
                        ],
                        default="userinfo_endpoint",
                        help_text="Indicates the source from which the user information claims should be extracted. This can be the ID token or the User Info endpoint.",
                        max_length=100,
                        verbose_name="user information claims extracted from",
                    ),
                ),
                (
                    "check_op_availability",
                    models.BooleanField(
                        default=False,
                        help_text="Whether to check OIDC Provider availability before doing the authentication request.",
                        verbose_name="check OIDC Provider availability",
                    ),
                ),
                (
                    "options",
                    django_jsonform.models.fields.JSONField(
                        default=dict,
                        help_text="Options relevant for a specific Identity Provider.",
                        verbose_name="options",
                    ),
                ),
                (
                    "oidc_provider",
                    models.ForeignKey(
                        help_text="Specifies which OIDC Provider to use.",
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="mozilla_django_oidc_db.oidcprovider",
                        verbose_name="OIDC Provider",
                    ),
                ),
            ],
            options={
                "verbose_name": "OIDC client",
                "verbose_name_plural": "OIDC clients",
            },
        ),
    ]



from urllib.parse import parse_qs, urlparse

from django.urls import reverse

import pytest
import requests

from mozilla_django_oidc_db.models import (
    OIDCClient,
    OIDCProvider,
    UserInformationClaimsSources,
)

from .utils import keycloak_login

KEYCLOAK_BASE_URL = "http://localhost:8080/realms/test/"


@pytest.mark.vcr
def test_client_id_secret_full_flow(
    keycloak_config, mock_state_and_nonce, client, django_user_model, vcr
):
    login_url = reverse("login-keycloak")
    django_login_response = client.get(login_url)
    assert django_login_response.status_code == 302

    # simulate login to Keycloak
    redirect_uri = keycloak_login(django_login_response["Location"])

    # complete the login flow on our end
    callback_response = client.get(redirect_uri)

    assert callback_response.status_code == 302
    assert callback_response["Location"] == "/admin/"

    # a user was created
    assert django_user_model.objects.count() == 1

    # check that the token request was performed as expected
    token_request = next(
        req
        for req in vcr.requests
        if req.uri == f"{KEYCLOAK_BASE_URL}protocol/openid-connect/token"
        and req.method == "POST"
    )
    assert token_request is not None
    assert b"client_id=testid" in token_request.body
    assert b"secret=7DB3KUAAizYCcmZufpHRVOcD0TOkNO3I" in token_request.body
    assert "Authorization" not in token_request.headers


@pytest.mark.vcr
def test_credentials_in_basic_auth_header(
    keycloak_config: OIDCClient,
    mock_state_and_nonce,
    client,
    django_user_model,
    vcr,
):
    assert keycloak_config.oidc_provider

    provider = keycloak_config.oidc_provider
    provider.oidc_token_use_basic_auth = True
    provider.save()

    django_login_response = client.get(reverse("login-keycloak"))
    # simulate login to Keycloak
    redirect_uri = keycloak_login(django_login_response["Location"])

    # complete the login flow on our end
    callback_response = client.get(redirect_uri)
    assert callback_response.status_code == 302
    assert callback_response["Location"] == "/admin/"

    # check that the token request was performed as expected
    token_request = next(
        req
        for req in vcr.requests
        if req.uri == f"{KEYCLOAK_BASE_URL}protocol/openid-connect/token"
        and req.method == "POST"
    )
    assert token_request is not None

    assert "Authorization" in token_request.headers
    bits = token_request.headers["Authorization"].split(" ")
    assert len(bits) == 2
    assert bits[0] == "Basic"

    assert b"client_id=testid" in token_request.body
    assert b"secret=" not in token_request.body


@pytest.mark.vcr
@pytest.mark.oidcconfig(
    # Set up client configured to return JWT from userinfo endpoint instead of plain
    # JSON. Credentials from ``docker/import`` realm export.
    oidc_rp_client_id="test-userinfo-jwt",
    oidc_rp_client_secret="ktGlGUELd1FR7dTXc84L7dJzUTjCtw9S",
    userinfo_claims_source=UserInformationClaimsSources.userinfo_endpoint,
)
def test_return_jwt_from_userinfo_endpoint(
    keycloak_config: OIDCClient,
    mock_state_and_nonce,
    client,
    django_user_model,
):
    django_login_response = client.get(reverse("login-keycloak"))
    # simulate login to Keycloak
    redirect_uri = keycloak_login(django_login_response["Location"])

    # complete the login flow on our end
    callback_response = client.get(redirect_uri)
    assert callback_response.status_code == 302
    assert callback_response["Location"] == "/admin/"

    # a user was created
    assert django_user_model.objects.count() == 1


@pytest.mark.vcr
@pytest.mark.oidcconfig(make_users_staff=True)
def test_session_refresh(
    keycloak_config,
    settings,
    mock_state_and_nonce,
    client,
    django_user_model,
    vcr,
    mocker,
):
    session = requests.Session()
    settings.MIDDLEWARE = settings.MIDDLEWARE + [
        "mozilla_django_oidc_db.middleware.SessionRefresh"
    ]
    settings.OIDC_RENEW_ID_TOKEN_EXPIRY_SECONDS = 60
    login_url = reverse("login-keycloak")

    django_login_response = client.get(login_url)
    # simulate login to Keycloak
    redirect_uri = keycloak_login(django_login_response["Location"], session=session)
    # complete the login flow on our end
    callback_response = client.get(redirect_uri)

    assert callback_response.status_code == 302
    assert callback_response["Location"] == "/admin/"

    admin_response = client.get("/admin/")

    # User was successfully logged in
    assert admin_response.status_code == 200

    # when the user refreshes the admin index, SessionRefresh should be called and
    # should redirect user to Keycloak
    mocker.patch("mozilla_django_oidc.middleware.time.time", lambda: 10**11)

    admin_response = client.get("/admin/")

    assert "/realms/test/protocol/openid-connect/auth" in admin_response["Location"]

    # Following the Keycloak response should redirect the user to the callback immediately,
    # because the user still has an active session with Keycloak
    keycloak_response = session.get(admin_response["Location"], allow_redirects=False)

    assert keycloak_response.status_code == 302
    assert "/oidc/callback/" in keycloak_response.headers["Location"]

    app_response = client.get(keycloak_response.headers["Location"])

    assert app_response.status_code == 302
    assert app_response.url == "/admin/"

    admin_response = client.get("/admin/")

    # User can reach the admin index again
    assert admin_response.status_code == 200



from urllib.parse import parse_qs, urlparse

from django.contrib.sessions.middleware import SessionMiddleware
from django.http.response import HttpResponseRedirect
from django.test import RequestFactory
from django.urls import reverse

import pytest

from mozilla_django_oidc_db.constants import CONFIG_IDENTIFIER_SESSION_KEY
from mozilla_django_oidc_db.middleware import SessionRefresh
from mozilla_django_oidc_db.models import OIDCClient

from .utils import create_or_update_configuration


@pytest.fixture(scope="session")
def dummy_view():
    def get_response(*args, **kwargs):
        pass

    return get_response


@pytest.fixture(scope="session")
def session_middleware(dummy_view):
    return SessionMiddleware(dummy_view)


@pytest.fixture(scope="session")
def session_refresh(dummy_view):
    return SessionRefresh(dummy_view)


@pytest.fixture()
def request_factory(rf: RequestFactory, session_middleware):
    def _factory(config_identifier: str = "", *, path: str = ""):
        request = rf.get(path or "/")
        session_middleware(request)
        session = request.session
        session[CONFIG_IDENTIFIER_SESSION_KEY] = config_identifier
        session.save()
        return request

    return _factory


@pytest.fixture()
def config_factory(db):
    def _factory(config_identifier: str, /, **overrides):
        BASE = f"https://mock-oidc-provider-{config_identifier}:9999"

        config = create_or_update_configuration(
            f"{config_identifier}-provider",
            config_identifier,
            {
                "enabled": True,
                "oidc_rp_client_id": "fake",
                "oidc_rp_client_secret": "even-faker",
                "oidc_rp_sign_algo": "RS256",
                "oidc_op_discovery_endpoint": f"{BASE}/oidc/",
                "oidc_op_jwks_endpoint": f"{BASE}/oidc/jwks",
                "oidc_op_authorization_endpoint": f"{BASE}/oidc/auth",
                "oidc_op_token_endpoint": f"{BASE}/oidc/token",
                "oidc_op_user_endpoint": f"{BASE}/oidc/user",
                **overrides,
            },
        )

        return config

    return _factory


@pytest.mark.oidcconfig(enabled=False)
def test_sessionrefresh_oidc_not_enabled(
    dummy_config: OIDCClient,
    request_factory,
    session_refresh: SessionRefresh,
):
    request = request_factory(dummy_config.identifier)

    # Running the middleware should return None, since OIDC is disabled
    result = session_refresh(request)

    assert result is None


@pytest.mark.oidcconfig(
    enabled=True,
    oidc_rp_client_id="initial-client-id",
    oidc_rp_scopes_list=["openid", "email"],
)
def test_sessionrefresh_config_always_refreshed(
    dummy_config: OIDCClient,
    request_factory,
    session_refresh: SessionRefresh,
    mocker,
):
    """
    Middleware should refresh the config on every call
    """
    mocker.patch.object(session_refresh, "is_refreshable_url", return_value=True)
    request = request_factory(dummy_config.identifier)

    result1 = session_refresh(request)
    assert isinstance(result1, HttpResponseRedirect)
    query1 = parse_qs(urlparse(result1.url).query)
    assert query1["client_id"] == ["initial-client-id"]
    assert query1["scope"] == ["openid email"]

    dummy_config.oidc_rp_client_id = "some-other-id"
    dummy_config.oidc_rp_scopes_list = ["openid", "email", "other_scope"]  # type: ignore
    dummy_config.save()

    result2 = session_refresh(request)
    assert isinstance(result2, HttpResponseRedirect)
    query2 = parse_qs(urlparse(result2.url).query)
    assert query2["client_id"] == ["some-other-id"]
    assert query2["scope"] == ["openid email other_scope"]


@pytest.mark.oidcconfig(enabled=True)
def test_sessionrefresh_config_use_defaults(
    dummy_config,
    settings,
    session_refresh: SessionRefresh,
    request_factory,
    mocker,
):
    """
    Middleware should respect fallbacks to settings/defaults.
    """
    settings.OIDC_AUTHENTICATION_CALLBACK_URL = "admin:index"
    mocker.patch.object(session_refresh, "is_refreshable_url", return_value=True)

    request = request_factory(dummy_config.identifier)

    result = session_refresh(request)

    assert isinstance(result, HttpResponseRedirect)
    query = parse_qs(urlparse(result.url).query)
    assert query["redirect_uri"] == ["http://testserver/admin/"]
    assert len(query["nonce"][0]) == 32  # default set on middleware dynamic_setting


@pytest.mark.django_db
@pytest.mark.parametrize(
    "config_identifier",
    (
        "test-oidc-not-configured",
        "test-oidc-another-not-configured",
    ),
)
def test_sessionfresh_selects_correct_backend_based_on_session_parameters(
    config_identifier,
    config_factory,
    request_factory,
    session_refresh: SessionRefresh,
    mocker,
):
    config_factory(
        "test-oidc-not-configured", enabled=True, oidc_rp_client_id="empty-config"
    )
    config_factory(
        "test-oidc-another-not-configured",
        enabled=True,
        oidc_rp_client_id="another-empty-config",
    )
    mocker.patch.object(session_refresh, "is_refreshable_url", return_value=True)
    request = request_factory(config_identifier)

    result1 = session_refresh(request)

    assert isinstance(result1, HttpResponseRedirect)
    assert f"https://mock-oidc-provider-{config_identifier}" in result1["Location"]
    query1 = parse_qs(urlparse(result1.url).query)
    config = OIDCClient.objects.get(identifier=config_identifier)

    assert query1["client_id"] == [config.oidc_rp_client_id]


@pytest.mark.django_db
@pytest.mark.parametrize(
    "config_identifier",
    (
        "test-oidc-not-configured",
        "test-oidc-another-not-configured",
    ),
)
def test_sessionfresh_adds_config_specific_callback_url_to_exempt_urls(
    config_identifier,
    config_factory,
    request_factory,
    session_refresh: SessionRefresh,
    mocker,
):
    class MockUser:
        @property
        def is_authenticated(self):
            return True

    config_factory("test-oidc-not-configured", enabled=True)
    config_factory(
        "test-oidc-another-not-configured",
        enabled=True,
    )

    callback_url = reverse("oidc_authentication_callback")

    request = request_factory(config_identifier, path=callback_url)
    request.user = MockUser()
    session_refresh._set_config_from_request(request)

    assert callback_url in session_refresh.exempt_urls
    assert session_refresh.is_refreshable_url(request) is False


@pytest.mark.django_db
def test_sessionfresh_does_nothing_for_non_oidc_requests(
    rf,
    session_refresh: SessionRefresh,
    mocker,
):
    mocker.patch.object(session_refresh, "is_refreshable_url", return_value=True)
    request = rf.get("/")

    result = session_refresh(request)

    assert result is None



import logging

import pytest

from mozilla_django_oidc_db.registry import register as registry

from .factories import UserFactory


@pytest.mark.oidcconfig(
    enabled=True,
    username_claim=["username"],
    extra_options={
        "user_settings.claim_mappings.first_name": [],
        "user_settings.claim_mappings.email": ["email"],
        "groups_settings.superuser_group_names": ["Superuser"],
        "groups_settings.claim_mapping": ["groups"],
    },
)
@pytest.mark.django_db
def test_update_user_from_claims(dummy_config):
    user = UserFactory.create(
        username="testuser",
        email="testuser@example.com",
        first_name="Test",
        is_superuser=False,
    )

    plugin = registry["test-oidc"]

    plugin.update_user(
        user,
        claims={
            "username": "testuser-MODIFIED",  # Username will not be modified
            "first_name": "Test-MODIFIED",  # First name claim path is not configured, so will not be updated
            "email": "testuser+MODIFIED@example.com",  # This should be updated
            "groups": ["Superuser"],
        },
    )

    user.refresh_from_db()

    assert user.username == "testuser"
    assert user.first_name == "Test"
    assert user.email == "testuser+MODIFIED@example.com"
    assert user.is_superuser


@pytest.mark.oidcconfig(
    enabled=True,
    username_claim=["username"],
    extra_options={
        "groups_settings.claim_mapping": ["not-present"],
    },
)
@pytest.mark.django_db
def test_no_groups_claim(dummy_config, caplog):
    user = UserFactory.create()

    plugin = registry["test-oidc"]

    with caplog.at_level(logging.DEBUG):
        plugin.update_user(
            user,
            claims={},
        )

    assert (
        "The configured groups claim 'not-present' was not found in the user info."
        in caplog.text
    )



# Generated by Django 5.0.4 on 2024-05-25 19:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("mozilla_django_oidc_db", "0002_migrate_to_claim_field"),
    ]

    operations = [
        migrations.AddField(
            model_name="openidconnectconfig",
            name="oidc_keycloak_idp_hint",
            field=models.CharField(
                blank=True,
                help_text="Specific for Keycloak: parameter that indicates which identity provider should be used (therefore skipping the Keycloak login screen).",
                max_length=1000,
                verbose_name="Keycloak Identity Provider hint",
            ),
        ),
        migrations.AddField(
            model_name="openidconnectconfig",
            name="oidc_op_logout_endpoint",
            field=models.URLField(
                blank=True,
                help_text="URL of your OpenID Connect provider logout endpoint",
                max_length=1000,
                verbose_name="Logout endpoint",
            ),
        ),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Iterable

from .constants import UNIQUE_PLUGIN_ID_MAX_LENGTH

if TYPE_CHECKING:
    from .plugins import OIDCPlugin


class OIDCRegistry:
    _registry: dict[str, OIDCPlugin]

    def __init__(self):
        self._registry = {}

    def __call__(
        self, unique_identifier: str
    ) -> Callable[[type[OIDCPlugin]], type[OIDCPlugin]]:

        if len(unique_identifier) > UNIQUE_PLUGIN_ID_MAX_LENGTH:
            raise ValueError(
                f"The unique identifier '{unique_identifier}' is longer than "
                f"{UNIQUE_PLUGIN_ID_MAX_LENGTH} characters."
            )

        def decorator(plugin_cls: type[OIDCPlugin]) -> type[OIDCPlugin]:
            if unique_identifier in self._registry:
                raise ValueError(
                    f"The unique identifier '{unique_identifier}' is already present "
                    "in the registry."
                )

            plugin = plugin_cls(identifier=unique_identifier)
            self._registry[unique_identifier] = plugin
            return plugin_cls

        return decorator

    def items(self) -> Iterable[tuple[str, OIDCPlugin]]:
        return self._registry.items()

    def __getitem__(self, key: str) -> OIDCPlugin:
        return self._registry[key]


# Sentinel to provide the registry. You can easily instantiate another
# :class:`Registry` object to use as dependency injection in tests.
register = OIDCRegistry()



from __future__ import annotations

import logging
from abc import abstractmethod
from typing import Any, Protocol, TypeAlias, runtime_checkable

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser, AnonymousUser
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.http import HttpRequest, HttpResponse

from glom import Path, glom

from .config import get_setting_from_config
from .constants import OIDC_ADMIN_CONFIG_IDENTIFIER
from .exceptions import MissingIdentifierClaim
from .models import OIDCClient
from .registry import register
from .schemas import ADMIN_OPTIONS_SCHEMA
from .typing import ClaimPath, JSONObject
from .utils import get_groups_by_name, obfuscate_claims
from .views import AdminCallbackView

logger = logging.getLogger(__name__)

missing = object()


class OIDCBasePluginProtocol(Protocol):
    identifier: str
    schema: JSONObject

    def get_config(self) -> OIDCClient: ...

    def get_setting(self, attr: str, *args) -> Any: ...

    @abstractmethod
    def get_schema(self) -> JSONObject:
        """Get the JSON schema of the ``options`` field on the :class:`~mozilla_django_oidc_db.models.OIDCClient` model."""
        ...

    @abstractmethod
    def validate_settings(self) -> None:
        """Check the validity of the settings in the provider and client configuration."""
        ...

    @abstractmethod
    def handle_callback(self, request: HttpRequest) -> HttpResponse:
        """Return an HttpResponse based on the callback view specified on the plugin.

        .. code:: python

           def view(self, request: HttpRequest) -> HttpResponse:
               return admin_callback_view(request)

        """
        ...

    @abstractmethod
    def get_extra_params(
        self, request: HttpRequest, extra_params: dict[str, str | bytes]
    ) -> dict[str, str | bytes]: ...


class BaseOIDCPlugin:
    def __init__(self, identifier: str):
        self.identifier = identifier

    def get_config(self) -> OIDCClient:
        return OIDCClient.objects.resolve(self.identifier)

    def get_setting(self, attr: str, *args) -> Any:
        config = self.get_config()

        return get_setting_from_config(config, attr, *args)

    def get_extra_params(
        self, request: HttpRequest, extra_params: dict[str, str | bytes]
    ) -> dict[str, str | bytes]:
        return extra_params


@runtime_checkable
class AnonymousUserOIDCPluginProtocol(OIDCBasePluginProtocol, Protocol):
    def get_or_create_user(
        self,
        access_token: str,
        id_token: str,
        payload: JSONObject,
        request: HttpRequest,
    ) -> AnonymousUser | None: ...


@runtime_checkable
class AbstractUserOIDCPluginProtocol(OIDCBasePluginProtocol, Protocol):
    def create_user(self, claims: JSONObject) -> AbstractUser: ...

    def update_user(self, user: AbstractUser, claims: JSONObject) -> AbstractUser: ...

    def filter_users_by_claims(
        self, claims: JSONObject
    ) -> models.Manager[AbstractUser]:
        """Return all users matching the specified subject."""
        ...

    def verify_claims(self, claims: JSONObject) -> bool:
        """Verify the provided claims to decide if authentication should be allowed."""
        ...


OIDCPlugin: TypeAlias = AbstractUserOIDCPluginProtocol | AnonymousUserOIDCPluginProtocol


admin_callback_view = AdminCallbackView.as_view()


@register(OIDC_ADMIN_CONFIG_IDENTIFIER)
class OIDCAdminPlugin(BaseOIDCPlugin, AbstractUserOIDCPluginProtocol):
    schema = ADMIN_OPTIONS_SCHEMA

    def verify_claims(self, claims: JSONObject) -> bool:
        """Verify the provided claims to decide if authentication should be allowed."""
        assert claims, "Empty claims should have been blocked earlier"

        config = self.get_config()

        sensitive_claims = config.options["user_settings"].get("sensitive_claims", [])
        # The identifier is considered sensitive by default
        sensitive_claims.append(
            config.options["user_settings"]["claim_mappings"]["username"]
        )

        obfuscated_claims = obfuscate_claims(claims, sensitive_claims)

        logger.debug("OIDC claims received: %s", obfuscated_claims)

        # check if we have an identifier
        try:
            self.get_username(claims, raise_on_empty=True)
        except MissingIdentifierClaim as exc:
            logger.error(
                "'%s' not in OIDC claims, cannot proceed with authentication",
                " > ".join(exc.claim_bits),
                exc_info=exc,
            )
            return False
        return True

    def get_username(self, claims: JSONObject, *, raise_on_empty: bool = False) -> str:
        """
        Given the claims, extract the username value.
        """

        config = self.get_config()

        claim_bits = config.options["user_settings"]["claim_mappings"]["username"]
        unique_id = glom(claims, Path(*claim_bits), default="")
        if raise_on_empty and not unique_id:
            raise MissingIdentifierClaim(claim_bits=claim_bits)
        return unique_id

    def validate_settings(self):
        # Equivalent of checks in upstream __init__ method
        config = self.get_config()

        if (
            config.oidc_rp_sign_algo.startswith("RS")
            or config.oidc_rp_sign_algo.startswith("ES")
        ) and (
            config.oidc_rp_idp_sign_key == ""
            and config.oidc_provider
            and config.oidc_provider.oidc_op_jwks_endpoint == ""
        ):
            msg = "{} alg requires OIDC_RP_IDP_SIGN_KEY or OIDC_OP_JWKS_ENDPOINT to be configured."
            raise ImproperlyConfigured(msg.format(config.oidc_rp_sign_algo))

    def filter_users_by_claims(
        self, claims: JSONObject
    ) -> models.Manager[AbstractUser]:
        """Return all users matching the specified subject."""
        UserModel = get_user_model()

        username = self.get_username(claims)
        assert username, (
            "A username must be provided. Did you forget to check empty usernames "
            "in the claim verification?"
        )

        lookup = UserModel.USERNAME_FIELD

        config = self.get_config()
        if not config.options["user_settings"].get("username_case_sensitive", True):
            lookup += "__iexact"

        return UserModel.objects.filter(**{lookup: username})

    def create_user(self, claims: JSONObject) -> AbstractUser:
        """Return object for a newly created user account."""
        username = self.get_username(claims)

        logger.debug("Creating Admin OIDC user: %s", username)

        UserModel = get_user_model()
        user = UserModel.objects.create_user(**{UserModel.USERNAME_FIELD: username})
        self.update_user(user, claims)
        return user

    def update_user(self, user: AbstractUser, claims: JSONObject) -> AbstractUser:
        """
        Update existing user with new claims, if necessary save, and return user.

        This method checks dynamic settings, which are provided by our
        :class:`~mozilla_django_oidc_db.models.OIDCClient` model. If the
        respective fields do not exist on the config model being used, you must make
        sure they exist as Django settings.
        """
        logger.debug("Updating Admin OIDC user %s", user)

        touched_fields: set[str] = set()

        config = self.get_config()

        for model_field, claim_bits in config.options["user_settings"][
            "claim_mappings"
        ].items():
            if model_field == "username":
                # We do not update the username
                continue

            # If no path is specified to a claim, skip it.
            if not claim_bits:
                continue

            value = glom(claims, Path(*claim_bits), default=missing)
            if value is missing:
                continue

            setattr(user, model_field, value)
            touched_fields.add(model_field)

        # Users can only be promoted to staff. Staff rights are never taken by OIDC.
        if config.options["groups_settings"]["make_users_staff"] and not user.is_staff:
            user.is_staff = True
            touched_fields.add("is_staff")

        groups_settings = config.options.get("groups_settings", {})
        groups = self._get_groups(claims, groups_settings.get("claim_mapping", []))

        # Assign superuser status to the user if the user is a member of at least one
        # specific group. Superuser status is explicitly removed if the user is not or
        # no longer member of at least one of these groups.
        if (
            superuser_group_names := groups_settings.get("superuser_group_names")
        ) and groups:
            # superuser is granted if there is overlap between the groups claim and the
            # specified superuser grup names
            make_superuser = bool(set(groups) & set(superuser_group_names))
            if make_superuser != user.is_superuser:
                user.is_superuser = make_superuser
                touched_fields.add("is_superuser")

        user.save(update_fields=touched_fields)

        self._set_user_groups(
            user,
            groups,
            default_group_names=groups_settings.get("default_groups", []),
            sync_missing_groups=groups_settings.get("sync", True),
            sync_groups_glob=groups_settings.get("sync_pattern", "*"),
        )

        return user

    def _get_groups(
        self, claims: JSONObject, groups_claim: ClaimPath
    ) -> list[str] | None:
        if not groups_claim:
            return

        # extract the specified claim
        groups: list[str] | str = glom(claims, Path(*groups_claim), default=missing)
        if groups is missing:
            logger.debug(
                "The configured groups claim '%s' was not found in the user info.",
                " > ".join(groups_claim),
            )
            groups = []

        # could be a string instead of a list of strings
        if isinstance(groups, str):
            groups = [groups]

        if not all(isinstance(name, str) for name in groups):
            logger.warning("Aborting! Groups is not a list of strings: %r", groups)
            return

        return groups

    def _set_user_groups(
        self,
        user: AbstractUser,
        groups: list[str] | None,
        default_group_names: list[str],
        sync_missing_groups: bool,
        sync_groups_glob: str,
    ) -> None:
        """
        Synchronize the groups from the config/claims and the model.
        """
        if groups is None:
            return

        # Incorporate the default groups into the desired state of groups.
        desired_group_names = set(default_group_names) | set(groups)

        # Update the user's group memberships, if required
        django_group_names = set(user.groups.values_list("name", flat=True))
        if set(desired_group_names) == django_group_names:
            return

        # Create missing groups if required
        existing_groups = get_groups_by_name(
            desired_group_names, sync_groups_glob, sync_missing_groups
        )

        # at this point, existing_groups is the full collection of groups that should be
        # set on the user model, because:
        #
        # * desired_group_names is equal to or a superset of the existing groups (superset
        #   if group synchronization is off or groups are excluded by the glob)
        # * the groups currently set on the user may contain groups not present in the
        #   desired group names, as the latter only reflects the default groups + groups in
        #   the claims
        user.groups.set(existing_groups)

    def get_schema(self) -> JSONObject:
        return ADMIN_OPTIONS_SCHEMA

    def handle_callback(self, request: HttpRequest) -> HttpResponse:
        return admin_callback_view(request)



from __future__ import annotations

from collections.abc import Sequence
from typing import Protocol, TypeAlias

from django.http import HttpRequest, HttpResponseBase

JSONPrimitive: TypeAlias = str | int | float | bool | None
JSONValue: TypeAlias = "JSONPrimitive | list[JSONValue] | JSONObject"
JSONObject: TypeAlias = dict[str, JSONValue]

ClaimPath: TypeAlias = Sequence[str]


class DjangoView(Protocol):
    def __call__(self, request: HttpRequest) -> HttpResponseBase: ...



from django.utils.translation import gettext_lazy as _

ADMIN_OPTIONS_SCHEMA = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "title": "Options",
    "description": _("OIDC Configuration options."),
    "type": "object",
    "additionalProperties": True,
    "properties": {
        "user_settings": {
            "description": _("Settings for managing the local Django Users."),
            "type": "object",
            "properties": {
                "claim_mappings": {
                    "description": _(
                        "Mapping between the Django User model fields and a path to a claim value."
                    ),
                    "type": "object",
                    "properties": {
                        "username": {
                            "description": _("Path to the claim to use as username."),
                            "type": "array",
                            "items": {
                                "type": "string",
                            },
                        },
                        "first_name": {
                            "description": _("Path to the claim to use as first name."),
                            "type": "array",
                            "items": {
                                "type": "string",
                            },
                        },
                        "last_name": {
                            "description": _("Path to the claim to use as last name."),
                            "type": "array",
                            "items": {
                                "type": "string",
                            },
                        },
                        "email": {
                            "description": _("Path to the claim to use as email."),
                            "type": "array",
                            "items": {
                                "type": "string",
                            },
                        },
                    },
                    "required": ["username"],
                    "additionalProperties": True,
                },
                "username_case_sensitive": {
                    "description": _(
                        "Whether the username of the Django User is case sensitive. Defaults to true."
                    ),
                    "type": "boolean",
                },
                "sensitive_claims": {
                    "description": _(
                        "List of paths to values in the claims which are considered sensitive. The username claim is considered sensitive by default."
                    ),
                    "type": "array",
                    "items": {
                        "type": "array",
                        "items": {
                            "type": "string",
                        },
                    },
                },
            },
            "required": ["claim_mappings"],
            "additionalProperties": True,
        },
        "groups_settings": {
            "description": _(
                "Settings required to assign the created users to the right Django groups."
            ),
            "type": "object",
            "properties": {
                "claim_mapping": {
                    "description": _(
                        "Path to the claim value that contains the groups that the user should be a member of."
                    ),
                    "type": "array",
                    "items": {"type": "string"},
                },
                "sync": {
                    "description": _(
                        "Whether local Django user groups should be created for group names present in the groups claim (if they do not exist)."
                    ),
                    "type": "boolean",
                },
                "sync_pattern": {
                    "description": _(
                        "The glob pattern that local groups must match to be synchronised to the local database."
                    ),
                    "type": "string",
                },
                "default_groups": {
                    "description": _(
                        "Names of the groups to which every user logging in with OIDC will be assigned."
                    ),
                    "type": "array",
                    "items": {"type": "string"},
                },
                "make_users_staff": {
                    "description": _(
                        "Users will be flagged as being a staff user automatically. This allows users to login to the admin interface. By default they have no permissions, even if they are staff."
                    ),
                    "type": "boolean",
                },
                "superuser_group_names": {
                    "description": _(
                        "If any of these group names are present in the claims upon login, the user will be marked as a superuser. If none of these groups are present the user will lose superuser permissions."
                    ),
                    "type": "array",
                    "items": {"type": "string"},
                },
            },
        },
    },
}



# Generated by Django 4.2.20 on 2025-04-04 08:26

from django.db import migrations
from glom import glom

from ..constants import (
    OIDC_ADMIN_CONFIG_IDENTIFIER,
    OIDC_ADMIN_PROVIDER_CONFIG_IDENTIFIER,
)


def move_data_backwards(apps, schema_editor):
    OIDCClient = apps.get_model("mozilla_django_oidc_db", "OIDCClient")
    OpenIDConnectConfig = apps.get_model(
        "mozilla_django_oidc_db", "OpenIDConnectConfig"
    )
    Group = apps.get_model("auth", "Group")

    # There can be multiple configurations, but we pick one
    new_config = OIDCClient.objects.filter(
        identifier=OIDC_ADMIN_CONFIG_IDENTIFIER
    ).first()
    if not new_config:
        return

    if not new_config.oidc_provider:
        # The model was likely never configured. We cannot migrate data backwards because
        # we don't have a fully configured client/provider.
        return

    username_claim_mapping = []
    if other_user_mappings := glom(
        new_config.options, "user_settings.claim_mappings", default={}
    ):
        username_claim_mapping = other_user_mappings["username"]
        del other_user_mappings["username"]

    group_settings = new_config.options.get("groups_settings", {})

    old_config = OpenIDConnectConfig.objects.create(
        enabled=new_config.enabled,
        # Provider settings
        oidc_op_discovery_endpoint=(
            new_config.oidc_provider.oidc_op_discovery_endpoint
        ),
        oidc_op_jwks_endpoint=new_config.oidc_provider.oidc_op_jwks_endpoint,
        oidc_op_authorization_endpoint=(
            new_config.oidc_provider.oidc_op_authorization_endpoint
        ),
        oidc_op_token_endpoint=new_config.oidc_provider.oidc_op_token_endpoint,
        oidc_op_user_endpoint=new_config.oidc_provider.oidc_op_user_endpoint,
        oidc_op_logout_endpoint=(new_config.oidc_provider.oidc_op_logout_endpoint),
        oidc_token_use_basic_auth=new_config.oidc_provider.oidc_token_use_basic_auth,
        oidc_use_nonce=new_config.oidc_provider.oidc_use_nonce,
        oidc_nonce_size=new_config.oidc_provider.oidc_nonce_size,
        oidc_state_size=new_config.oidc_provider.oidc_state_size,
        # Client settings
        oidc_rp_client_id=new_config.oidc_rp_client_id,
        oidc_rp_client_secret=new_config.oidc_rp_client_secret,
        oidc_rp_sign_algo=new_config.oidc_rp_sign_algo,
        oidc_rp_scopes_list=new_config.oidc_rp_scopes_list,
        oidc_rp_idp_sign_key=new_config.oidc_rp_idp_sign_key,
        oidc_keycloak_idp_hint=new_config.oidc_keycloak_idp_hint,
        userinfo_claims_source=new_config.userinfo_claims_source,
        username_claim=username_claim_mapping,
        claim_mapping=other_user_mappings,
        groups_claim=group_settings.get("claim_mapping", []),
        sync_groups=group_settings.get("sync", True),
        sync_groups_glob_pattern=group_settings.get("sync_pattern", "*"),
        make_users_staff=group_settings.get("make_users_staff", False),
        superuser_group_names=group_settings.get("superuser_group_names", []),
    )

    groups = [
        Group.objects.get_or_create(name=group_name)[0]
        for group_name in group_settings.get("default_groups", [])
    ]

    old_config.default_groups.set(groups)
    old_config.save()


def move_data_forward(apps, schema_editor):
    OIDCClient = apps.get_model("mozilla_django_oidc_db", "OIDCClient")
    OIDCProvider = apps.get_model("mozilla_django_oidc_db", "OIDCProvider")
    OpenIDConnectConfig = apps.get_model(
        "mozilla_django_oidc_db", "OpenIDConnectConfig"
    )

    # Only one solo model should be present
    old_config = OpenIDConnectConfig.objects.first()
    if not old_config:
        return

    oidc_provider = OIDCProvider.objects.create(
        identifier=OIDC_ADMIN_PROVIDER_CONFIG_IDENTIFIER,
        oidc_op_discovery_endpoint=(old_config.oidc_op_discovery_endpoint),
        oidc_op_jwks_endpoint=old_config.oidc_op_jwks_endpoint,
        oidc_op_authorization_endpoint=(old_config.oidc_op_authorization_endpoint),
        oidc_op_token_endpoint=old_config.oidc_op_token_endpoint,
        oidc_op_user_endpoint=old_config.oidc_op_user_endpoint,
        oidc_op_logout_endpoint=old_config.oidc_op_logout_endpoint,
        oidc_token_use_basic_auth=old_config.oidc_token_use_basic_auth,
        oidc_use_nonce=old_config.oidc_use_nonce,
        oidc_nonce_size=old_config.oidc_nonce_size,
        oidc_state_size=old_config.oidc_state_size,
    )

    OIDCClient.objects.create(
        identifier=OIDC_ADMIN_CONFIG_IDENTIFIER,
        enabled=old_config.enabled,
        oidc_provider=oidc_provider,
        oidc_rp_client_id=old_config.oidc_rp_client_id,
        oidc_rp_client_secret=old_config.oidc_rp_client_secret,
        oidc_rp_sign_algo=old_config.oidc_rp_sign_algo,
        oidc_rp_scopes_list=old_config.oidc_rp_scopes_list,
        oidc_rp_idp_sign_key=old_config.oidc_rp_idp_sign_key,
        oidc_keycloak_idp_hint=old_config.oidc_keycloak_idp_hint,
        userinfo_claims_source=old_config.userinfo_claims_source,
        options={
            "user_settings": {
                "claim_mappings": {
                    "username": old_config.username_claim,
                    **old_config.claim_mapping,
                },
                "username_case_sensitive": True,
            },
            "groups_settings": {
                "claim_mapping": old_config.groups_claim,
                "sync": old_config.sync_groups,
                "sync_pattern": old_config.sync_groups_glob_pattern,
                "default_groups": [
                    group.name for group in old_config.default_groups.all()
                ],
                "make_users_staff": old_config.make_users_staff,
                "superuser_group_names": old_config.superuser_group_names,
            },
        },
    )


class Migration(migrations.Migration):

    dependencies = [
        (
            "mozilla_django_oidc_db",
            "0006_oidcprovider_oidcclient",
        ),
    ]

    operations = [
        migrations.RunPython(move_data_forward, move_data_backwards),
    ]



# Generated by Django 4.2.20 on 2025-04-11 14:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("mozilla_django_oidc_db", "0007_migrate_oidc_data"),
    ]

    operations = [
        migrations.DeleteModel(
            name="OpenIDConnectConfig",
        ),
        migrations.CreateModel(
            name="OpenIDConnectConfig",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "managed": False,
            },
        ),
    ]



from django.contrib.auth.signals import user_logged_in
from django.dispatch import receiver

from .constants import CONFIG_IDENTIFIER_SESSION_KEY
from .registry import register as registry


@receiver([user_logged_in], dispatch_uid="oidcdb.set_config_identifier")
def set_oidcdb_config_identifier_on_session(sender, user, request, **kwargs):
    """
    Record the OIDC config identifier on the session, this is needed so the callback view
    can retrieve the config in case of a SessionRefresh flow.
    """
    if hasattr(user, "_oidcdb_config_identifier"):
        request.session[CONFIG_IDENTIFIER_SESSION_KEY] = user._oidcdb_config_identifier


def populate_oidc_config_models(sender, **kwargs) -> None:
    """
    See which OIDC plugins are registered and make sure that the corresponding configuration model exists in the
    database.
    """
    if not (apps := kwargs.get("apps")):
        return

    OIDCClient = apps.get_model("mozilla_django_oidc_db", "OIDCClient")

    for unique_identifier, plugin in registry.items():
        OIDCClient.objects.get_or_create(identifier=unique_identifier)



from django.apps import apps
from django.core.checks import Error, Warning

import pytest

from mozilla_django_oidc_db.checks import check_authenticate_class, check_callback_class


@pytest.fixture(scope="session")
def app_configs():
    app_config = apps.get_app_config(app_label="mozilla_django_oidc_db")
    return [app_config]


def test_nothing_reported_when_checking_other_app_label(settings):
    settings.OIDC_AUTHENTICATE_CLASS = object()  # this is invalid
    app_config = apps.get_app_config(app_label="admin")

    messages = check_authenticate_class(app_configs=[app_config])

    assert len(messages) == 0


def test_check_authenticate_class_ok(app_configs, settings):
    settings.OIDC_AUTHENTICATE_CLASS = (
        "mozilla_django_oidc_db.views.OIDCAuthenticationRequestView"
    )

    messages = check_authenticate_class(app_configs=app_configs)

    assert len(messages) == 0


def test_check_authenticate_class_not_a_string(app_configs, settings):
    settings.OIDC_AUTHENTICATE_CLASS = object()

    messages = check_authenticate_class(app_configs=app_configs)

    assert len(messages) == 1
    assert messages[0] == Error(
        "'settings.OIDC_AUTHENTICATE_CLASS' must be a string that can be imported.",
        hint=(
            "Use 'mozilla_django_oidc_db.views.OIDCAuthenticationRequestView' or a "
            "subclass of 'mozilla_django_oidc_db.views.OIDCAuthenticationRequestInitView'."
        ),
        id="mozilla_django_oidc_db.E001",
    )


def test_check_authenticate_class_invalid_view(app_configs, settings):
    settings.OIDC_AUTHENTICATE_CLASS = "django.views.View"

    messages = check_authenticate_class(app_configs=app_configs)

    assert len(messages) == 1
    assert messages[0] == Warning(
        "'settings.OIDC_AUTHENTICATE_CLASS' should be a subclass of 'OIDCAuthenticationRequestInitView'.",
        hint=(
            "Use 'mozilla_django_oidc_db.views.OIDCAuthenticationRequestView' or a "
            "subclass of 'mozilla_django_oidc_db.views.OIDCAuthenticationRequestInitView'."
        ),
        id="mozilla_django_oidc_db.W001",
    )


def test_check_callback_class_ok(app_configs, settings):
    settings.OIDC_CALLBACK_CLASS = "mozilla_django_oidc_db.views.OIDCCallbackView"

    messages = check_callback_class(app_configs=app_configs)

    assert len(messages) == 0


def test_check_callback_class_not_a_string(app_configs, settings):
    settings.OIDC_CALLBACK_CLASS = object()

    messages = check_callback_class(app_configs=app_configs)

    assert len(messages) == 1
    assert messages[0] == Error(
        "'settings.OIDC_CALLBACK_CLASS' must be a string that can be imported.",
        hint=(
            "Use 'mozilla_django_oidc_db.views.OIDCCallbackView' or a "
            "subclass of it."
        ),
        id="mozilla_django_oidc_db.E002",
    )


def test_check_callback_class_invalid_view(app_configs, settings):
    settings.OIDC_CALLBACK_CLASS = "django.views.View"

    messages = check_callback_class(app_configs=app_configs)

    assert len(messages) == 1
    assert messages[0] == Warning(
        "'settings.OIDC_CALLBACK_CLASS' should be a subclass of 'OIDCInit'.",
        hint=(
            "Use 'mozilla_django_oidc_db.views.OIDCCallbackView' or a "
            "subclass of it."
        ),
        id="mozilla_django_oidc_db.W002",
    )



import warnings

from django.core.exceptions import ObjectDoesNotExist

from django_setup_configuration.configuration import BaseConfigurationStep
from django_setup_configuration.exceptions import ConfigurationRunFailed

from mozilla_django_oidc_db.forms import OIDCProviderForm
from mozilla_django_oidc_db.models import OIDCClient, OIDCProvider
from mozilla_django_oidc_db.setup_configuration.models import (
    AdminOIDCConfigurationModel,
    AdminOIDCConfigurationModelItem,
    OIDCConfigProviderModel,
    OIDCDiscoveryProviderConfig,
)


# TODO: We now have a single step that supports the yaml files used in versions <= 0.23.0 and >0.23.0.
# When we drop compatibility, this should be refactor to have two steps: one for the provider and
# one for the client models.
class AdminOIDCConfigurationStep(BaseConfigurationStep[AdminOIDCConfigurationModel]):
    """
    Configure the necessary settings to enable OpenID Connect authentication for admin users.

    This allows admin users to log in with Single Sign On (SSO) to access the management interface.
    """

    verbose_name = "Configuration for admin login via OpenID Connect"
    config_model = AdminOIDCConfigurationModel
    namespace = "oidc_db_config_admin_auth"
    enable_setting = "oidc_db_config_enable"

    def execute(self, model: AdminOIDCConfigurationModel) -> None:
        # Configure OIDC Providers
        for provider_config_model in model.providers:
            self._create_or_update_providers(provider_config_model)

        for config_model in model.items:
            self._create_or_update_configuration(config_model)

    def _get_endpoints(self, endpoint_config) -> dict:
        endpoints = {}
        if isinstance(endpoint_config, OIDCDiscoveryProviderConfig):
            endpoints.update(
                oidc_op_discovery_endpoint=endpoint_config.oidc_op_discovery_endpoint,
            )
        else:
            endpoints.update(
                oidc_op_authorization_endpoint=endpoint_config.oidc_op_authorization_endpoint,
                oidc_op_token_endpoint=endpoint_config.oidc_op_token_endpoint,
                oidc_op_user_endpoint=endpoint_config.oidc_op_user_endpoint,
                oidc_op_logout_endpoint=endpoint_config.oidc_op_logout_endpoint,
                oidc_op_jwks_endpoint=endpoint_config.oidc_op_jwks_endpoint,
            )
        return endpoints

    def _create_or_update_provider_deprecated(
        self, config_model: AdminOIDCConfigurationModelItem
    ) -> OIDCProvider:
        warnings.warn(
            "Specifying the OIDC Provider settings directly in the OIDC configuration is deprecated. "
            "Provide the settings for the OIDC Provider separately.",
            DeprecationWarning,
        )

        identifier = f"{config_model.identifier}-provider"
        settings_provider = {
            "identifier": identifier,
            "oidc_token_use_basic_auth": config_model.oidc_token_use_basic_auth,
            "oidc_use_nonce": config_model.oidc_use_nonce,
            "oidc_nonce_size": config_model.oidc_nonce_size,
            "oidc_state_size": config_model.oidc_state_size,
            **self._get_endpoints(config_model.endpoint_config),
        }

        provider, _ = OIDCProvider.objects.update_or_create(identifier=identifier)
        form = OIDCProviderForm(
            instance=provider,
            data=settings_provider,
        )
        if not form.is_valid():
            raise ConfigurationRunFailed(
                "Admin OIDC configuration field validation failed",
                form.errors.as_json(),
            )
        provider = form.save()
        return provider

    def _create_or_update_configuration(
        self, config_model: AdminOIDCConfigurationModelItem
    ) -> None:
        if not config_model.oidc_provider_identifier:
            provider = self._create_or_update_provider_deprecated(config_model)
        else:
            try:
                provider = OIDCProvider.objects.get(
                    identifier=config_model.oidc_provider_identifier
                )
            except ObjectDoesNotExist as exc:
                raise ConfigurationRunFailed(
                    f"Could not find an existing OIDC Provider with "
                    f"identifier `{config_model.oidc_provider_identifier}`."
                )

        all_settings = {
            "enabled": config_model.enabled,
            "oidc_rp_client_id": config_model.oidc_rp_client_id,
            "oidc_rp_client_secret": config_model.oidc_rp_client_secret,
            "oidc_rp_sign_algo": config_model.oidc_rp_sign_algo,
            "oidc_rp_scopes_list": config_model.oidc_rp_scopes_list,
            "oidc_rp_idp_sign_key": config_model.oidc_rp_idp_sign_key,
            "oidc_keycloak_idp_hint": config_model.oidc_keycloak_idp_hint,
            "userinfo_claims_source": config_model.userinfo_claims_source,
            "oidc_provider": provider,
            "options": {},
        }

        config, _ = OIDCClient.objects.update_or_create(
            identifier=config_model.identifier, defaults=all_settings
        )

        if config_model.options:
            config.options = config_model.options
        else:
            warnings.warn(
                "The OIDC configuration attributes ``username_claim``, ``claim_mapping``, ``sync_groups``, "
                "``sync_groups_glob_pattern``, ``make_users_staff``, ``superuser_group_names`` and ``default_groups`` "
                "are deprecated. Use the ``options`` attribute instead.",
                DeprecationWarning,
            )
            config.options = {
                "user_settings": {
                    "claim_mappings": {
                        "username": config_model.username_claim,
                        **config_model.claim_mapping,
                    }
                },
                "group_settings": {
                    "claim_mapping": config_model.groups_claim,
                    "sync": config_model.sync_groups,
                    "sync_pattern": config_model.sync_groups_glob_pattern,
                    "make_users_staff": config_model.make_users_staff,
                    "superuser_group_names": config_model.superuser_group_names,
                    "default_groups": config_model.default_groups,
                },
            }
        config.save()

    def _create_or_update_providers(
        self, provider_config_model: OIDCConfigProviderModel
    ) -> None:
        provider, _ = OIDCProvider.objects.update_or_create(
            identifier=provider_config_model.identifier
        )
        form = OIDCProviderForm(
            instance=provider,
            data={
                "identifier": provider_config_model.identifier,
                "oidc_token_use_basic_auth": provider_config_model.oidc_token_use_basic_auth,
                "oidc_use_nonce": provider_config_model.oidc_use_nonce,
                "oidc_nonce_size": provider_config_model.oidc_nonce_size,
                "oidc_state_size": provider_config_model.oidc_state_size,
                **self._get_endpoints(provider_config_model.endpoint_config),
            },
        )
        if not form.is_valid():
            raise ConfigurationRunFailed(
                "Admin OIDC configuration field validation failed",
                form.errors.as_json(),
            )
        form.save()



from __future__ import annotations

import logging
from typing import Any, cast

from django.contrib.auth import get_user_model
from django.contrib.auth.models import (
    AbstractBaseUser,
    AbstractUser,
    AnonymousUser,
)
from django.db import models
from django.http import HttpRequest

import requests
from mozilla_django_oidc.auth import OIDCAuthenticationBackend as BaseBackend
from typing_extensions import override

from .config import dynamic_setting, lookup_config
from .exceptions import MissingInitialisation
from .jwt import verify_and_decode_token
from .models import OIDCClient, UserInformationClaimsSources
from .plugins import (
    AbstractUserOIDCPluginProtocol,
    AnonymousUserOIDCPluginProtocol,
)
from .registry import register as registry
from .typing import JSONObject
from .utils import extract_content_type

logger = logging.getLogger(__name__)


missing = object()


class OIDCAuthenticationBackend(BaseBackend):
    """
    Custom backend modifying the upstream package behaviour.

    This backend looks up the configuration to use, which is set
    on the request by the init view. It does this by grabbing the state query parameter and looking
    in the session for this key. This key was set by the authentication request view.
    Scoped inside the state, there is the identifier of the configuration to use.

    The authenticate method saves the request on the backend instance.

    No configuration is loaded in :meth:`__init__` at all, instead we define properties
    to dynamically look this up. Django instantiates backends *a lot*, e.g. during
    permission checks. We only support the :meth:`authenticate` entrypoint like the
    upstream library.
    """

    request: HttpRequest | None = None  # set during the authenticate call
    config: OIDCClient | None = None  # set during the authenticate call

    # These should be functionaly equivalent to
    # :class:`mozilla_django_oidc.auth.OIDCAuthenticationBackend`.
    OIDC_OP_TOKEN_ENDPOINT = dynamic_setting[str]()
    OIDC_OP_USER_ENDPOINT = dynamic_setting[str]()
    OIDC_OP_JWKS_ENDPOINT = dynamic_setting[str | None](default=None)
    OIDC_RP_CLIENT_ID = dynamic_setting[str]()
    OIDC_RP_CLIENT_SECRET = dynamic_setting[str]()
    OIDC_RP_SIGN_ALGO = dynamic_setting[str](default="HS256")
    OIDC_RP_IDP_SIGN_KEY = dynamic_setting[str | None](default=None)

    @override
    def __init__(self, *args, **kwargs) -> None:
        # Deliberately empty, we discard all initialization from the parent class which
        # requires a config_class to be set. Even if we set it (declaratively), this is
        # not viable because it performs DB/cache IO to look up the config instance,
        # which would happen as well when Django goes through the auth backends for
        # permission checks.
        #
        # See https://github.com/maykinmedia/mozilla-django-oidc-db/issues/30

        # django-stubs returns AbstractBaseUser, but we depend on properties of
        # AbstractUser.
        self.UserModel = cast(AbstractUser, get_user_model())

    @override
    def get_settings(self, attr: str, *args: Any) -> Any:  # type: ignore
        """
        Override the upstream library get_settings.

        Upstream is django-settings based, and we store configuration in database
        records instead. We look up the configuration from the DB and check if the
        requested setting is defined there or not. If not, it is taken from the Django
        settings.
        """
        if not self.config:
            raise MissingInitialisation(
                "The configuration must be loaded from the authenticate entrypoint. It looks like "
                "you're trying to access configuration before this was called."
            )

        plugin = registry[self.config.identifier]
        plugin.validate_settings()

        return plugin.get_setting(attr, *args)

    def _check_candidate_backend(self) -> bool:
        return self.get_settings("enabled")

    # The method signature is checked by django when selecting a suitable backend. Our
    # signature is more strict than the upstream library. Check the upstream
    # `OIDCAuthenticationCallbackView` for the `auth.authenticate(**kwargs)` call if this
    # needs updating.
    @override
    def authenticate(  # type: ignore
        self,
        request: HttpRequest | None,
        nonce: str | None = None,
        code_verifier: str | None = None,
    ) -> AnonymousUser | AbstractBaseUser | None:
        """
        Authenticate the user with OIDC *iff* the conditions are met.

        Return ``None`` to skip to the next backend, raise
        :class:`django.core.exceptions.PermissionDenied` to stop in our tracks. Return
        a user object (real or anonymous) to signify success.
        """
        # if we don't get a request, we can't check anything, so skip to the next
        # backend. We need to grab the state and code from request.GET for OIDC.
        if request is None:
            return None

        self.config = lookup_config(request)
        self.request = request

        # Check if this backend should be considered to authenticate the user.
        is_candidate = self._check_candidate_backend()
        if not is_candidate:
            return None

        # Allright, now try to actually authenticate the user.
        user = super().authenticate(request, nonce=nonce, code_verifier=code_verifier)

        # Store the config class identifier on the user, so that we can store this in the user's
        # session after they have been successfully authenticated (by listening to the `user_logged_in` signal)
        if user:
            user._oidcdb_config_identifier = self.config.identifier  # type: ignore

        return user

    @override
    def verify_claims(self, claims: JSONObject) -> bool:
        """Verify the provided claims to decide if authentication should be allowed."""
        assert claims, "Empty claims should have been blocked earlier"
        assert self.config

        plugin = registry[self.config.identifier]
        assert isinstance(plugin, AbstractUserOIDCPluginProtocol)
        return plugin.verify_claims(claims)

    @override
    def get_userinfo(
        self, access_token: str, id_token: str, payload: JSONObject
    ) -> JSONObject:
        """
        Extract the user information, configurable whether to use the ID token or
        the userinfo endpoint for this
        """
        assert self.config

        if self.config.userinfo_claims_source == UserInformationClaimsSources.id_token:
            logger.debug("Extracting user information from ID token")
            return payload

        logger.debug("Retrieving user information from userinfo endpoint")

        # copy of upstream get_userinfo which doesn't support application/jwt yet.
        # Overridden to handle application/jwt responses.
        # See https://github.com/mozilla/mozilla-django-oidc/issues/517
        #
        # Specifying the preferred format in the ``Accept`` header does not work with
        # Keycloak, as it depends on the client settings.
        assert self.config.oidc_provider
        user_response = requests.get(
            self.config.oidc_provider.oidc_op_user_endpoint,
            headers={
                "Authorization": "Bearer {0}".format(access_token),
            },
            verify=self.get_settings("OIDC_VERIFY_SSL", True),
            timeout=self.get_settings("OIDC_TIMEOUT", None),
            proxies=self.get_settings("OIDC_PROXY", None),
        )
        user_response.raise_for_status()

        # From https://openid.net/specs/openid-connect-core-1_0.html#UserInfoResponse
        #
        # > The UserInfo Endpoint MUST return a content-type header to indicate which
        # > format is being returned.
        content_type = extract_content_type(user_response.headers["Content-Type"])
        match content_type:
            case "application/json":
                # the default case of upstream library
                return user_response.json()
            case "application/jwt":
                token = user_response.content
                # get the key from the configured keys endpoint
                # XXX: tested with asymmetric encryption. algorithms like HS256 rely on
                # out-of-band key exchange and are currently not supported until such a
                # case arrives.
                key = self.retrieve_matching_jwk(token)
                payload = verify_and_decode_token(token, key)
                return payload
            case _:
                raise ValueError(
                    f"Got an invalid Content-Type header value ({content_type}) "
                    "according to OpenID Connect Core 1.0 standard. Contact your "
                    "vendor."
                )

    @override
    def filter_users_by_claims(self, claims: JSONObject) -> models.Manager[AbstractUser]:  # type: ignore (parent function returns UserManager which is more specific than Manager)
        assert self.config
        plugin = registry[self.config.identifier]

        assert isinstance(plugin, AbstractUserOIDCPluginProtocol)
        return plugin.filter_users_by_claims(claims)

    @override
    def create_user(self, claims: JSONObject) -> AbstractUser:
        """Create an authenticated user."""
        assert self.config
        plugin = registry[self.config.identifier]

        assert isinstance(plugin, AbstractUserOIDCPluginProtocol)
        return plugin.create_user(claims)

    @override
    def update_user(self, user: AbstractUser, claims: JSONObject):
        assert self.config
        plugin = registry[self.config.identifier]

        assert isinstance(plugin, AbstractUserOIDCPluginProtocol)
        return plugin.update_user(user, claims)

    @override
    def get_or_create_user(self, access_token: str, id_token: str, payload: JSONObject) -> AnonymousUser | AbstractUser | None:  # type: ignore (parent function returns only an AbstractUser | None)
        """Get or create a user based on the tokens received."""
        assert self.config

        plugin = registry[self.config.identifier]
        assert isinstance(self.request, HttpRequest)

        if isinstance(plugin, AnonymousUserOIDCPluginProtocol):
            return plugin.get_or_create_user(
                access_token, id_token, payload, self.request
            )

        return super().get_or_create_user(access_token, id_token, payload)



from __future__ import annotations

from django.core.exceptions import (
    BadRequest,
    MultipleObjectsReturned,
)
from django.db import models
from django.utils.translation import gettext_lazy as _

from django_jsonform.models.fields import ArrayField, JSONField
from typing_extensions import deprecated

from .constants import UNIQUE_PLUGIN_ID_MAX_LENGTH
from .registry import register as registry
from .typing import JSONObject


@deprecated(
    "Left here so that proxy models that in the migrations have "
    'bases=("mozilla_django_oidc_db.openidconnectconfig",) can still run their migrations.'
)
class OpenIDConnectConfig(models.Model):
    class Meta:
        managed = False


def get_options_schema(instance: "OIDCClient") -> JSONObject:
    plugin = registry[instance.identifier]
    return plugin.get_schema()


class UserInformationClaimsSources(models.TextChoices):
    userinfo_endpoint = "userinfo_endpoint", _("Userinfo endpoint")
    id_token = "id_token", _("ID token")


def get_default_scopes() -> list[str]:
    """
    Returns the default scopes to request for OpenID Connect logins
    """
    return ["openid", "email", "profile"]


def get_claim_mapping() -> dict[str, list[str]]:
    # Map (some) claim names from https://openid.net/specs/openid-connect-core-1_0.html#Claims
    # to corresponding field names on the User model
    return {
        "email": ["email"],
        "first_name": ["given_name"],
        "last_name": ["family_name"],
    }


class OIDCProvider(models.Model):
    """
    Manage the configuration/connection parameters for a single OIDC Provider (OP).
    """

    identifier = models.SlugField(
        unique=True,
        verbose_name=_("identifier"),
        help_text=_("Unique identifier of the OIDC provider."),
        max_length=255,
    )

    oidc_op_discovery_endpoint = models.URLField(
        _("Discovery endpoint"),
        max_length=1000,
        help_text=_(
            "URL of your provider discovery endpoint ending with a slash "
            "(`.well-known/...` will be added automatically). "
            "If this is provided, the remaining endpoints can be omitted, as "
            "they will be derived from this endpoint."
        ),
        blank=True,
    )
    oidc_op_jwks_endpoint = models.URLField(
        _("JSON Web Key Set endpoint"),
        max_length=1000,
        help_text=_(
            "URL of your provider JSON Web Key Set endpoint. "
            "Required if `RS256` is used as signing algorithm."
        ),
        blank=True,
    )
    oidc_op_authorization_endpoint = models.URLField(
        _("Authorization endpoint"),
        max_length=1000,
        help_text=_("URL of your provider authorization endpoint"),
    )
    oidc_op_token_endpoint = models.URLField(
        _("Token endpoint"),
        max_length=1000,
        help_text=_("URL of your provider token endpoint"),
    )
    oidc_op_user_endpoint = models.URLField(
        _("User endpoint"),
        max_length=1000,
        help_text=_("URL of your provider userinfo endpoint."),
    )
    oidc_op_logout_endpoint = models.URLField(
        _("Logout endpoint"),
        max_length=1000,
        help_text=_("URL of your provider logout endpoint."),
        blank=True,
    )

    # Advanced settings
    oidc_token_use_basic_auth = models.BooleanField(
        _("Use Basic auth for token endpoint"),
        default=False,
        help_text=_(
            "If enabled, the client ID and secret are sent in the HTTP Basic auth "
            "header when obtaining the access token. Otherwise, they are sent in the "
            "request body.",
        ),
        blank=True,
    )
    oidc_use_nonce = models.BooleanField(
        _("Use nonce"),
        help_text=_("Controls whether the client uses nonce verification"),
        default=True,
        blank=True,
    )
    oidc_nonce_size = models.PositiveIntegerField(
        _("Nonce size"),
        help_text=_("Sets the length of the random string used for nonce verification"),
        default=32,
        blank=True,
    )
    oidc_state_size = models.PositiveIntegerField(
        _("State size"),
        help_text=_("Sets the length of the random string used for state verification"),
        default=32,
        blank=True,
    )

    class Meta:
        verbose_name = _("OIDC Provider")
        verbose_name_plural = _("OIDC Providers")

    def __str__(self):
        return _("OIDC Provider {identifier}").format(identifier=self.identifier)


class OIDCClientManager(models.Manager):
    def resolve(self, identifier: str) -> OIDCClient:
        try:
            return OIDCClient.objects.select_related("oidc_provider").get(
                identifier=identifier
            )
        except (OIDCClient.DoesNotExist, MultipleObjectsReturned) as exc:
            raise BadRequest("Could not look up the referenced configuration.") from exc


class OIDCClient(models.Model):
    """
    Hold the client configuration for the Relying Party (RP).

    At minimum, the client credentials need to be configured for the associated OIDC Provider (OP).
    Additional dynamic configuration options can be specified that are tied to specific
    use cases.
    """

    identifier = models.SlugField(
        unique=True,
        verbose_name=_("identifier"),
        help_text=_("Unique identifier for the client."),
        max_length=UNIQUE_PLUGIN_ID_MAX_LENGTH,
    )

    enabled = models.BooleanField(
        _("enabled"),
        default=False,
        help_text=_(
            "The client must be enabled before users can authenticate through it."
        ),
    )
    oidc_provider = models.ForeignKey(
        to=OIDCProvider,
        on_delete=models.PROTECT,
        verbose_name=_("OIDC Provider"),
        help_text=_("Specifies which OIDC Provider to use."),
        # Needed so that we can create empty models at startup
        null=True,
    )

    oidc_rp_client_id = models.CharField(
        _("Client ID"),
        max_length=1000,
        help_text=_("Client ID provided by the OIDC Provider"),
    )
    oidc_rp_client_secret = models.CharField(
        _("Secret"),
        max_length=1000,
        help_text=_("Secret provided by the OIDC Provider"),
    )
    oidc_rp_scopes_list = ArrayField(
        verbose_name=_("Scopes"),
        base_field=models.CharField(_("Scope"), max_length=50),
        default=get_default_scopes,
        blank=True,
        help_text=_("Scopes that are requested during login"),
    )
    oidc_rp_sign_algo = models.CharField(
        _("OpenID sign algorithm"),
        max_length=50,
        help_text=_("Algorithm the Identity Provider uses to sign ID tokens"),
        default="RS256",
    )
    oidc_rp_idp_sign_key = models.CharField(
        _("Sign key"),
        max_length=1000,
        help_text=_(
            "Key the Identity Provider uses to sign ID tokens in the case of an RSA sign algorithm. "
            "Should be the signing key in PEM or DER format."
        ),
        blank=True,
    )

    # Keycloak specific config
    oidc_keycloak_idp_hint = models.CharField(
        _("Keycloak Identity Provider hint"),
        max_length=1000,
        help_text=_(
            "Specific for Keycloak: parameter that indicates which identity provider "
            "should be used (therefore skipping the Keycloak login screen)."
        ),
        blank=True,
    )

    userinfo_claims_source = models.CharField(
        verbose_name=_("user information claims extracted from"),
        choices=UserInformationClaimsSources.choices,
        max_length=100,
        default=UserInformationClaimsSources.userinfo_endpoint,
        help_text=_(
            "Indicates the source from which the user information claims should be extracted. This can be the ID token or the User Info endpoint."
        ),
    )

    # Additional settings
    check_op_availability = models.BooleanField(
        verbose_name=_("check OIDC Provider availability"),
        default=False,
        help_text=_(
            "Whether to check OIDC Provider availability before doing the authentication request."
        ),
    )

    # Customisable options
    options = JSONField(
        _("options"),
        help_text=_("Options relevant for a specific Identity Provider."),
        schema=get_options_schema,
        default=dict,
    )

    objects = OIDCClientManager()

    class Meta:
        verbose_name = _("OIDC client")
        verbose_name_plural = _("OIDC clients")

    def __str__(self):
        return _("OIDC client {identifier}").format(identifier=self.identifier)

    @property
    def oidc_rp_scopes(self) -> str:
        """
        Scopes should be formatted as a string with spaces
        """
        return " ".join(self.oidc_rp_scopes_list)  # type: ignore



import factory

from ..models import OIDCClient, OIDCProvider


class OIDCProviderFactory(factory.django.DjangoModelFactory):
    identifier = factory.Faker("word")

    oidc_op_discovery_endpoint = factory.Faker("url")

    oidc_op_jwks_endpoint = factory.LazyAttribute(
        lambda provider: f"{provider.oidc_op_discovery_endpoint}protocol/openid-connect/certs"
    )
    oidc_op_authorization_endpoint = factory.LazyAttribute(
        lambda provider: f"{provider.oidc_op_discovery_endpoint}openid-connect/auth"
    )
    oidc_op_token_endpoint = factory.LazyAttribute(
        lambda provider: f"{provider.oidc_op_discovery_endpoint}protocol/openid-connect/token"
    )
    oidc_op_user_endpoint = factory.LazyAttribute(
        lambda provider: f"{provider.oidc_op_discovery_endpoint}protocol/openid-connect/userinfo"
    )
    oidc_op_logout_endpoint = factory.LazyAttribute(
        lambda provider: f"{provider.oidc_op_discovery_endpoint}protocol/openid-connect/logout"
    )

    class Meta:
        model = OIDCProvider
        django_get_or_create = ("identifier",)


class OIDCClientFactory(factory.django.DjangoModelFactory):
    oidc_provider = factory.SubFactory(OIDCProviderFactory)
    identifier = factory.Faker("word")
    oidc_rp_client_id = factory.Faker("word")
    oidc_rp_client_secret = factory.Faker("word")
    options = factory.Dict({"bla": "bla"})

    class Meta:
        model = OIDCClient
        django_get_or_create = ("identifier",)

    class Params:
        with_admin_options = factory.Trait(
            options=factory.Dict(
                {
                    "user_settings": {
                        "claim_mappings": {
                            "username": ["sub"],
                            "email": ["email"],
                            "last_name": [],
                            "first_name": [],
                        },
                        "username_case_sensitive": False,
                        "sensitive_claims": [],
                    },
                    "groups_settings": {
                        "make_users_staff": True,
                        "sync": False,
                        "default_groups": [],
                        "claim_mapping": [],
                        "superuser_group_names": [],
                        "sync_pattern": "*",
                    },
                }
            )
        )



# Generated by Django 4.2.9 on 2024-05-01 15:32

from django.db import migrations, models

import django_jsonform.models.fields

import mozilla_django_oidc_db.models


class Migration(migrations.Migration):

    dependencies = [
        ("auth", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="OpenIDConnectConfig",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "enabled",
                    models.BooleanField(
                        default=False,
                        help_text="Indicates whether OpenID Connect for authentication/authorization is enabled",
                        verbose_name="enable",
                    ),
                ),
                (
                    "oidc_rp_client_id",
                    models.CharField(
                        help_text="OpenID Connect client ID provided by the OIDC Provider",
                        max_length=1000,
                        verbose_name="OpenID Connect client ID",
                    ),
                ),
                (
                    "oidc_rp_client_secret",
                    models.CharField(
                        help_text="OpenID Connect secret provided by the OIDC Provider",
                        max_length=1000,
                        verbose_name="OpenID Connect secret",
                    ),
                ),
                (
                    "oidc_rp_sign_algo",
                    models.CharField(
                        default="HS256",
                        help_text="Algorithm the Identity Provider uses to sign ID tokens",
                        max_length=50,
                        verbose_name="OpenID sign algorithm",
                    ),
                ),
                (
                    "oidc_rp_scopes_list",
                    django_jsonform.models.fields.ArrayField(
                        base_field=models.CharField(
                            max_length=50, verbose_name="OpenID Connect scope"
                        ),
                        blank=True,
                        default=mozilla_django_oidc_db.models.get_default_scopes,
                        help_text="OpenID Connect scopes that are requested during login",
                        size=None,
                        verbose_name="OpenID Connect scopes",
                    ),
                ),
                (
                    "oidc_op_jwks_endpoint",
                    models.URLField(
                        blank=True,
                        help_text="URL of your OpenID Connect provider JSON Web Key Set endpoint. Required if `RS256` is used as signing algorithm.",
                        max_length=1000,
                        verbose_name="JSON Web Key Set endpoint",
                    ),
                ),
                (
                    "oidc_op_authorization_endpoint",
                    models.URLField(
                        help_text="URL of your OpenID Connect provider authorization endpoint",
                        max_length=1000,
                        verbose_name="Authorization endpoint",
                    ),
                ),
                (
                    "oidc_op_token_endpoint",
                    models.URLField(
                        help_text="URL of your OpenID Connect provider token endpoint",
                        max_length=1000,
                        verbose_name="Token endpoint",
                    ),
                ),
                (
                    "oidc_op_user_endpoint",
                    models.URLField(
                        help_text="URL of your OpenID Connect provider userinfo endpoint",
                        max_length=1000,
                        verbose_name="User endpoint",
                    ),
                ),
                (
                    "oidc_rp_idp_sign_key",
                    models.CharField(
                        blank=True,
                        help_text="Key the Identity Provider uses to sign ID tokens in the case of an RSA sign algorithm. Should be the signing key in PEM or DER format.",
                        max_length=1000,
                        verbose_name="Sign key",
                    ),
                ),
                (
                    "oidc_op_discovery_endpoint",
                    models.URLField(
                        blank=True,
                        help_text="URL of your OpenID Connect provider discovery endpoint ending with a slash (`.well-known/...` will be added automatically). If this is provided, the remaining endpoints can be omitted, as they will be derived from this endpoint.",
                        max_length=1000,
                        verbose_name="Discovery endpoint",
                    ),
                ),
                (
                    "claim_mapping",
                    models.JSONField(
                        default=mozilla_django_oidc_db.models.get_claim_mapping,
                        help_text="Mapping from user-model fields to OIDC claims",
                        verbose_name="claim mapping",
                    ),
                ),
                (
                    "groups_claim",
                    models.CharField(
                        blank=True,
                        default="roles",
                        help_text="The name of the OIDC claim that holds the values to map to local user groups.",
                        max_length=50,
                        verbose_name="groups claim",
                    ),
                ),
                (
                    "make_users_staff",
                    models.BooleanField(
                        default=False,
                        help_text="Users will be flagged as being a staff user automatically. This allows users to login to the admin interface. By default they have no permissions, even if they are staff.",
                        verbose_name="make users staff",
                    ),
                ),
                (
                    "sync_groups",
                    models.BooleanField(
                        default=True,
                        help_text="If checked, local user groups will be created for group names present in the groups claim, if they do not exist yet locally.",
                        verbose_name="Create local user groups if they do not exist yet",
                    ),
                ),
                (
                    "sync_groups_glob_pattern",
                    models.CharField(
                        default="*",
                        help_text="The glob pattern that groups must match to be synchronized to the local database.",
                        max_length=255,
                        verbose_name="groups glob pattern",
                    ),
                ),
                (
                    "username_claim",
                    models.CharField(
                        default="sub",
                        help_text="The name of the OIDC claim that is used as the username",
                        max_length=50,
                        verbose_name="username claim",
                    ),
                ),
                (
                    "oidc_exempt_urls",
                    django_jsonform.models.fields.ArrayField(
                        base_field=models.CharField(
                            max_length=1000, verbose_name="Exempt URL"
                        ),
                        blank=True,
                        default=list,
                        help_text="This is a list of absolute url paths, regular expressions for url paths, or Django view names. This plus the mozilla-django-oidc urls are exempted from the session renewal by the SessionRefresh middleware.",
                        size=None,
                        verbose_name="URLs exempt from session renewal",
                    ),
                ),
                (
                    "oidc_nonce_size",
                    models.PositiveIntegerField(
                        default=32,
                        help_text="Sets the length of the random string used for OpenID Connect nonce verification",
                        verbose_name="Nonce size",
                    ),
                ),
                (
                    "oidc_state_size",
                    models.PositiveIntegerField(
                        default=32,
                        help_text="Sets the length of the random string used for OpenID Connect state verification",
                        verbose_name="State size",
                    ),
                ),
                (
                    "oidc_use_nonce",
                    models.BooleanField(
                        default=True,
                        help_text="Controls whether the OpenID Connect client uses nonce verification",
                        verbose_name="Use nonce",
                    ),
                ),
                (
                    "default_groups",
                    models.ManyToManyField(
                        blank=True,
                        help_text="The default groups to which every user logging in with OIDC will be assigned",
                        to="auth.group",
                        verbose_name="default groups",
                    ),
                ),
                (
                    "userinfo_claims_source",
                    models.CharField(
                        choices=[
                            ("userinfo_endpoint", "Userinfo endpoint"),
                            ("id_token", "ID token"),
                        ],
                        default="userinfo_endpoint",
                        help_text="Indicates the source from which the user information claims should be extracted.",
                        max_length=100,
                        verbose_name="user information claims extracted from",
                    ),
                ),
                (
                    "superuser_group_names",
                    django_jsonform.models.fields.ArrayField(
                        base_field=models.CharField(
                            max_length=50, verbose_name="Superuser group name"
                        ),
                        blank=True,
                        default=list,
                        help_text="If any of these group names are present in the claims upon login, the user will be marked as a superuser. If none of these groups are present the user will lose superuser permissions.",
                        size=None,
                        verbose_name="Superuser group names",
                    ),
                ),
                (
                    "oidc_token_use_basic_auth",
                    models.BooleanField(
                        default=False,
                        help_text="If enabled, the client ID and secret are sent in the HTTP Basic auth header when obtaining the access token. Otherwise, they are sent in the request body.",
                        verbose_name="Use Basic auth for token endpoint",
                    ),
                ),
            ],
            options={
                "verbose_name": "OpenID Connect configuration",
            },
        ),
    ]



from django.contrib.auth.models import User
from django.db import IntegrityError
from django.http import HttpRequest
from django.test import Client
from django.urls import reverse
from django.utils.translation import gettext as _

import pytest
from pytest_django.asserts import assertContains, assertRedirects

from mozilla_django_oidc_db.models import (
    OIDCClient,
    UserInformationClaimsSources,
)
from mozilla_django_oidc_db.typing import JSONObject
from mozilla_django_oidc_db.views import OIDCAuthenticationRequestInitView
from testapp.backends import MockBackend

from .factories import UserFactory


@pytest.fixture
def mock_auth_backend(request, mocker):
    marker = request.node.get_closest_marker("mock_backend_claims")
    claims: JSONObject = marker.args[0] if marker else {"sub": "some_username"}
    mock_backend = MockBackend(claims=claims)
    backend_path = f"{MockBackend.__module__}.{MockBackend.__qualname__}"
    mocker.patch(
        "django.contrib.auth._get_backends", return_value=[(mock_backend, backend_path)]
    )
    return mock_backend


@pytest.fixture
def callback_client(callback_request: HttpRequest, client: Client) -> Client:
    session = client.session
    for key, value in callback_request.session.items():
        session[key] = value
    session.save()

    return client


@pytest.mark.mock_backend_claims(
    {"email": "collision@example.com", "sub": "some_username"}
)
@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
)
def test_duplicate_email_unique_constraint_violated(
    dummy_config: OIDCClient,
    callback_request: HttpRequest,
    callback_client: Client,
    mock_auth_backend: MockBackend,
    mocker,
):
    UserFactory.create(username="nonmatchingusername", email="collision@example.com")
    mocker.patch.object(
        mock_auth_backend,
        "create_user",
        side_effect=IntegrityError(
            """duplicate key value violates unique constraint "filled_email_unique"""
            """"\nDETAIL:  Key (email)=(collision@example.com) already exists.\n"""
        ),
    )
    callback_url = reverse("oidc_authentication_callback")
    error_url = reverse("admin-oidc-error")

    # check that the response redirects to the error page
    callback_response = callback_client.get(callback_url, {**callback_request.GET})

    assertRedirects(callback_response, error_url)

    # check that the error page displays error information
    error_page = callback_client.get(error_url)

    assert error_page.status_code == 200
    expected_error = (
        """duplicate key value violates unique constraint "filled_email_unique"""
        """"\nDETAIL:  Key (email)=(collision@example.com) already exists.\n"""
    )
    assert error_page.context["oidc_error"] == expected_error
    assertContains(error_page, "duplicate key value violates unique constraint")


@pytest.mark.mock_backend_claims(
    {
        "sub": "some_username",
        "email": "collision@example.com",
        "wrong_is_superuser_value_type": "",  # should be boolean instead of string
    }
)
@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    extra_options={
        "user_settings.claim_mappings.is_superuser": ["wrong_is_superuser_value_type"]
    },
)
def test_validation_error_during_authentication(
    dummy_config: OIDCClient,
    callback_request: HttpRequest,
    callback_client: Client,
    mock_auth_backend: MockBackend,
):
    UserFactory.create(username="some_username", email="admin@example.com")
    callback_url = reverse("oidc_authentication_callback")
    error_url = reverse("admin-oidc-error")

    # check that the response redirects to the error page
    callback_response = callback_client.get(callback_url, {**callback_request.GET})

    assertRedirects(callback_response, error_url)

    # check that the error page displays error information
    error_page = callback_client.get(error_url)

    assert error_page.status_code == 200
    expected_error = _("“%(value)s” value must be either True or False.") % {
        "value": ""
    }
    assert error_page.context["oidc_error"] == expected_error
    assertContains(error_page, expected_error)


@pytest.mark.mock_backend_claims(
    {
        "email": "nocollision@example.com",
        "sub": "some_username",
    }
)
@pytest.mark.auth_request(next="/admin/")
@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
)
def test_happy_flow(
    dummy_config: OIDCClient,
    callback_request: HttpRequest,
    callback_client: Client,
    mock_auth_backend: MockBackend,
):
    callback_url = reverse("oidc_authentication_callback")

    callback_response = callback_client.get(callback_url, {**callback_request.GET})

    assertRedirects(callback_response, "/admin/", fetch_redirect_response=False)
    user = User.objects.get(email="nocollision@example.com")
    assert user.username == "some_username"


@pytest.mark.mock_backend_claims({})
@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
)
def test_empty_claims_returned(
    dummy_config: OIDCClient,
    callback_request: HttpRequest,
    callback_client: Client,
    mock_auth_backend: MockBackend,
):
    """
    Assert that the login procedure fails gracefully when no user claims are returned.
    """
    callback_url = reverse("oidc_authentication_callback")
    error_url = reverse("admin-oidc-error")

    callback_response = callback_client.get(callback_url, {**callback_request.GET})

    assertRedirects(callback_response, error_url, fetch_redirect_response=False)
    assert not User.objects.exists()


def test_tampering_with_parameters_invalid_state(
    dummy_config: OIDCClient,
    callback_request: HttpRequest,
    callback_client: Client,
):
    callback_url = reverse("oidc_authentication_callback")
    valid_state_key = callback_request.GET["state"]
    assert valid_state_key != "invalid-state"

    callback_response = callback_client.get(callback_url, {"state": "invalid-state"})

    assert callback_response.status_code == 400


def test_invalid_reference_to_config_identifier(
    dummy_config: OIDCClient,
    callback_request: HttpRequest,
    callback_client: Client,
):
    # This only guards against programmer mistakes downstream that would lead to the
    # end-user being able to manipulate the config_identifier being set in the state. It is
    # a synthetic case.
    callback_url = reverse("oidc_authentication_callback")
    state_key = callback_request.GET["state"]
    session = callback_client.session
    del session["oidc_states"][state_key]["config_identifier"]
    session.save()

    callback_response = callback_client.get(callback_url, {"state": state_key})

    assert callback_response.status_code == 400


@pytest.mark.django_db
@pytest.mark.oidcconfig(enabled=False, oidc_op_authorization_endpoint="bad")
def test_wrong_config_model_used(
    dummy_config: OIDCClient,
    auth_request: HttpRequest,
    client: Client,
):
    init_view = OIDCAuthenticationRequestInitView.as_view(identifier="test-oidc")
    init_view(auth_request)
    # there is only one state expected
    state_key = list(auth_request.session["oidc_states"].keys())[0]
    callback_url = reverse("oidc_authentication_callback")
    session = client.session
    for key in auth_request.session.keys():
        session[key] = auth_request.session[key]
    session.save()

    callback_response = client.get(
        callback_url, {"state": state_key, "code": "irrelevant"}
    )

    assert callback_response.status_code == 302
    assert callback_response.url == "/admin/login/failure/"


@pytest.mark.auth_request(next="/admin/")
@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
)
def test_error_first_cleared_after_succesful_login(
    dummy_config: OIDCClient,
    callback_request: HttpRequest,
    callback_client: Client,
    mock_auth_backend: MockBackend,
):
    session = callback_client.session
    session["oidc-error"] = "some error"
    session.save()
    callback_url = reverse("oidc_authentication_callback")
    error_url = reverse("admin-oidc-error")

    # check that error page works
    error_response = callback_client.get(error_url)
    assert error_response.status_code == 200

    # now on succesfull login, the error must be cleared
    callback_response = callback_client.get(callback_url, {**callback_request.GET})

    assertRedirects(callback_response, "/admin/", fetch_redirect_response=False)

    error_response = callback_client.get(error_url)
    assert error_response.status_code == 403



from django.test import Client
from django.urls import reverse

import pytest
from requests import Session

from mozilla_django_oidc_db.models import OIDCClient

from .utils import keycloak_login


@pytest.mark.vcr
@pytest.mark.oidcconfig(make_users_staff=True)
def test_use_config_class_from_state_over_config_class_from_session(
    keycloak_config: OIDCClient,
    dummy_config: OIDCClient,
    mock_state_and_nonce,
    client: Client,
):
    """
    When using two different OIDC configs, ensure that their state doesn't get mixed up.

    First, we authenticate in the django admin, this is the config that uses the
    session refresh, and the config set up through fixtures.

    Second, we have another OIDC config that uses another provider. The state of the
    first authentication may not affect the second authentication flow.
    """
    session = Session()
    # login to the admin
    login_url = reverse("login-keycloak")
    django_login_response = client.get(login_url)
    redirect_uri = keycloak_login(django_login_response["Location"], session=session)
    callback_response = client.get(redirect_uri, follow=True)
    # sanity check
    assert callback_response.wsgi_request.path == reverse("admin:index")

    # set up an authentication flow & state with another config - all the credentials
    # are otherwise the same - the only difference is where the callback redirects after
    # succesful authentication
    login_url2 = reverse("login-keycloak-custom")
    django_login_response2 = client.get(login_url2)
    # we expect to still be authenticated in the keycloak session, so we can fetch the
    # URL directly - and perform a sanity check!
    _response = session.get(django_login_response2["Location"], allow_redirects=False)
    redirect_uri2 = _response.headers["Location"]
    assert redirect_uri2.startswith("http://testserver/")
    callback_response2 = client.get(redirect_uri2, follow=True)
    assert callback_response2.wsgi_request.path == "/custom-success-url"



import logging
from typing import Any
from urllib.parse import parse_qs, urlsplit

from django.contrib import admin
from django.core.exceptions import DisallowedRedirect, PermissionDenied, ValidationError
from django.db import IntegrityError, transaction
from django.http import HttpRequest, HttpResponseBase, HttpResponseRedirect
from django.urls import reverse_lazy
from django.utils.http import url_has_allowed_host_and_scheme
from django.views import View
from django.views.generic import TemplateView

import requests
from mozilla_django_oidc.views import (
    OIDCAuthenticationCallbackView as BaseOIDCCallbackView,
    OIDCAuthenticationRequestView as BaseOIDCAuthRequestInitView,
)

from .config import get_setting_from_config, lookup_config, store_config
from .constants import OIDC_ADMIN_CONFIG_IDENTIFIER
from .exceptions import OIDCProviderOutage
from .models import OIDCClient
from .registry import register as registry

logger = logging.getLogger(__name__)

_OIDC_ERROR_SESSION_KEY = "oidc-error"
"""
Session key where to store authentication error messages.

During the callback flow, if any errors are encountered, they are stored in the session
under this key so that :class:`AdminLoginFailure` can read and display them to the
end-user.
"""

_RETURN_URL_SESSION_KEY = "oidc-db_redirect_next"
"""
Session key for the "next" URL to redirect the user to.

This is the equivalent of the "oidc_login_next" session key from mozilla_django_oidc,
which we deliberately do not rely on as their usage may change and it is private API.

In some situations the value of this session key needs to be used as base to properly
display problems (used in the ``failure_url`` flow of the callback view).
"""


def get_exception_message(exc: Exception) -> str:
    if isinstance(exc, ValidationError):
        # ValidationError can be raised as part of django.db.models.fields.Field.to_python,
        # and unfortunately we don't have any context about the exact field that raised
        # the exception.
        return exc.messages[0]
    return exc.args[0]


class OIDCCallbackView(View):
    """
    Route to the appropriate callback request handler.

    When a callback request is received, the state contains information about the
    configuration model to use. A particular configuration may require
    certain view behaviour. This view acts as a centralized entrypoint so that there
    is only a single callback endpoint required. It ensures that the configuration
    is extracted from the request.
    """

    def get(self, request: HttpRequest, *args, **kwargs) -> HttpResponseBase:
        """
        Extract the state from the request parameters and persist it.

        The state is extracted so that it's available for the authentication backend(s)
        and the downstream views.
        """
        store_config(request)
        configuration = lookup_config(request)
        plugin = registry[configuration.identifier]
        return plugin.handle_callback(request)


class OIDCAuthenticationCallbackView(BaseOIDCCallbackView):
    """
    Base callback view that retrieves the settings from the config object.
    """

    def get_settings(self, attr: str, *args: Any) -> Any:  # type: ignore
        """
        Look up the request setting from the database config.

        For the duration of the request, the configuration instance is cached on the
        view.
        """
        if (configuration := getattr(self, "_config", None)) is None:
            configuration = lookup_config(self.request)
            self._config = configuration
        return get_setting_from_config(configuration, attr, *args)


class AdminCallbackView(OIDCAuthenticationCallbackView):
    """
    Intercept errors raised by the authentication backend and display them.
    """

    failure_url = reverse_lazy("admin-oidc-error")

    def get(self, request: HttpRequest, *args, **kwargs):
        try:
            # ensure errors don't lead to half-created users
            with transaction.atomic():
                response = super().get(request)
        except (IntegrityError, ValidationError) as exc:
            logger.exception(
                "Something went wrong while attempting to authenticate via OIDC",
                exc_info=exc,
            )
            exc_message = get_exception_message(exc)
            request.session[_OIDC_ERROR_SESSION_KEY] = exc_message
            return self.login_failure()
        else:
            if _OIDC_ERROR_SESSION_KEY in request.session:
                del request.session[_OIDC_ERROR_SESSION_KEY]
        return response


default_callback_view = OIDCAuthenticationCallbackView.as_view()
admin_callback_view = AdminCallbackView.as_view()


class AdminLoginFailure(TemplateView):
    """
    Template view in admin style to display OIDC login errors
    """

    template_name = "admin/oidc_failure.html"

    def dispatch(self, request, *args, **kwargs):
        if _OIDC_ERROR_SESSION_KEY not in request.session:
            raise PermissionDenied()
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(admin.site.each_context(self.request))
        context["oidc_error"] = self.request.session[_OIDC_ERROR_SESSION_KEY]
        return context


class OIDCAuthenticationRequestInitView(BaseOIDCAuthRequestInitView):
    """
    A 'view' to start an OIDC authentication flow.

    This view class is parametrized with the identifier of the config model, so that
    the specific configuration can be retrieved and settings such as the identity provider endpoint
    to redirect the user to can be obtained.

    This view is not necessarily meant to be exposed directly via a URL pattern, but
    rather specific views are to be created from it, e.g.:

    .. code-block:: python

        >>> digid_init = OIDCAuthenticationRequestInitView.as_view(identifier="digid-oidc")
        >>> redirect_response = digid_init(request)
        # Redirect to some keycloak instance, for example.

    The ``__init__`` method of the Django ``View`` will add attributes to the view for
    any args/kwargs passed during init. So the view will have an attribute ``identifier``.
    Note that it only does it for attributes that are already defined as class attributes.

    These concrete views are intended to be wrapped by your own views so that you can
    supply the ``return_url`` parameter:

    .. code-block:: python

        def my_digid_login(request):
            return digid_init(request, return_url=request.GET["next"])

    Compared to :class:`mozilla_django_oidc.views.OIDCAuthenticationRequestView`, some
    extra actions are performed:

    * Any Keycloak IdP hint is added, if configured
    * The ``return_url`` is validated against unsafe redirects
    * The availability of the identity provider endpoint can be checked, if it's not
      available, the :class:`mozilla_django_oidc_db.exceptions.OIDCProviderOutage`
      exception is raised. Note that your own code needs to handle this appropriately!
    """

    identifier: str = ""
    """
    The identifier of the config model to get the settings from.

    Specify this as a kwarg in the ``as_view(identifier=...)`` class method.
    """

    allow_next_from_query: bool = True
    """
    Specify if the url-to-redirect-to may be provided as a query string parameter.

    For OIDC auth in the admin, you want to enable this to make URLs like
    ``/oidc/authenticate/?next=/admin/`` work as expected. For more advanced flows,
    you may want explicit control over this URL via your own wrapper view:

    .. code-block:: python

        digid_init = OIDCAuthenticationRequestInitView.as_view(
            identifier="config-identifier", allow_next_from_query=False
        )

        def my_digid_login(request):
            return digid_init(request, return_url="/some-fixed-url")
    """

    def get(
        self, request: HttpRequest, return_url: str = "", *args, **kwargs
    ) -> HttpResponseRedirect:
        if not self.allow_next_from_query:
            self._validate_return_url(request, return_url=return_url)

        if self.get_settings("check_op_availability", False):
            self.check_idp_availability()

        response = super().get(request, *args, **kwargs)

        # update the return_url value with what the upstream library extracted from the
        # GET query parameters.
        if self.allow_next_from_query:
            return_url = request.session["oidc_login_next"]

        # We add our own key to keep track of the redirect URL. In the case of
        # authentication failure (or canceled logins), the session is cleared by the
        # upstream library, so in the callback view we store this URL so that we know
        # where to redirect with the error information.
        request.session[_RETURN_URL_SESSION_KEY] = return_url

        # mozilla-django-oidc grabs this from request.GET and since that is not mutable,
        # it's easiest to just override the session key with the correct value.
        request.session["oidc_login_next"] = return_url

        # Store the config identifier to use in the state. We can not simply pass this as
        # a querystring parameter appended to redirect_uri, as these are likely to be
        # strictly validated. We must grab the state from the redirect Location.
        # This config reference is later used in the authentication callback view and
        # the authentication backend.
        query = parse_qs(urlsplit(response.url).query)
        state_params: list[str] = query["state"]
        assert len(state_params) == 1, "Expected only a single state parameter"
        state_key = state_params[0]

        # update the state. the parent class caused the session to be marked as modified,
        # so django's middleware will take care of persisting this to the session backend.
        state = request.session["oidc_states"][state_key]
        state["config_identifier"] = self.identifier

        return response

    def get_settings(self, attr: str, *args: Any) -> Any:  # type: ignore
        """
        Look up the request setting from the database config.

        For the duration of the request, the configuration instance is cached on the
        view.
        """
        if (config := getattr(self, "_config", None)) is None:
            config = OIDCClient.objects.get(identifier=self.identifier)
            self._config = config
        return get_setting_from_config(config, attr, *args)

    @staticmethod
    def _validate_return_url(request: HttpRequest, return_url: str) -> None:
        """
        Validate that the return URL meets the requirements.

        1. A non-empty value needs to be provided.
        2. The URL must be a safe redirect - only internal redirects are allowed.
        """
        if not return_url:
            raise ValueError("You must pass a return URL")

        url_is_safe = url_has_allowed_host_and_scheme(
            url=return_url,
            allowed_hosts=request.get_host(),
            require_https=request.is_secure(),
        )
        if not url_is_safe:
            raise DisallowedRedirect(f"Can't redirect to '{return_url}'")

    def check_idp_availability(self) -> None:
        """
        Hook for subclasses.

        Raise :class:`OIDCProviderOutage` if the Identity Provider is not available,
        which your application code needs to handle.

        The default implementation checks if the endpoint has a status code < 401.
        """
        endpoint = self.get_settings("OIDC_OP_AUTH_ENDPOINT")
        try:
            # Verify that the identity provider endpoint can be reached. This is where
            # the user ultimately gets redirected to.
            #
            # 5 seconds wait time is probably already too long for a good user
            # experience, but we don't want to be *too* aggressive.
            response = requests.get(endpoint, timeout=5.0)
            # some IDPs have been observed to return HTTP 400 because of the missing
            # query params
            if response.status_code > 400:
                response.raise_for_status()
        except requests.RequestException as exc:
            logger.info(
                "OIDC provider endpoint '%s' could not be retrieved",
                endpoint,
                exc_info=exc,
            )
            raise OIDCProviderOutage("Identity provider appears to be down.") from exc

    def get_extra_params(self, request: HttpRequest) -> dict[str, str]:
        """
        Add a keycloak identity provider hint if configured.
        """
        extra = super().get_extra_params(request)

        if kc_idp_hint := self.get_settings("OIDC_KEYCLOAK_IDP_HINT", ""):
            extra["kc_idp_hint"] = kc_idp_hint

        plugin = registry[self.identifier]
        extra = plugin.get_extra_params(request, extra)

        return extra


class OIDCAuthenticationRequestView(OIDCAuthenticationRequestInitView):
    """
    Start an OIDC authentication flow.

    This view is pre-configured to use the OIDC configuration included in this library,
    intended for admin authentication. Enable it in your Django settings with:

    .. code-block:: python

        OIDC_AUTHENTICATE_CLASS = (
            "mozilla_django_oidc_db.views.OIDCAuthenticationRequestView"
        )
    """

    identifier = OIDC_ADMIN_CONFIG_IDENTIFIER
    allow_next_from_query = True



"""
Test the OIDC Authenticaton Request flow with our custom views.
"""

from urllib.parse import parse_qs, urlsplit

from django.http import HttpRequest
from django.urls import reverse

import pytest

from mozilla_django_oidc_db.constants import OIDC_ADMIN_CONFIG_IDENTIFIER
from mozilla_django_oidc_db.exceptions import OIDCProviderOutage
from mozilla_django_oidc_db.plugins import OIDCAdminPlugin
from mozilla_django_oidc_db.registry import register
from mozilla_django_oidc_db.tests.factories import OIDCClientFactory
from mozilla_django_oidc_db.views import OIDCAuthenticationRequestInitView


@pytest.mark.oidcconfig(
    oidc_op_authorization_endpoint="http://localhost:8080/openid-connect/auth"
)
def test_default_config_flow(filled_admin_config, client):
    start_url = reverse("oidc_authentication_init")
    assert start_url == "/oidc/authenticate/"

    response = client.get(start_url, {"next": "/admin/"})

    # check that the view redirects to the identity provider
    assert response.status_code == 302
    parsed_url = urlsplit(response.url)
    assert parsed_url.scheme == "http"
    assert parsed_url.netloc == "localhost:8080"
    assert parsed_url.path == "/openid-connect/auth"

    # introspect state
    state_key = parse_qs(parsed_url.query)["state"][0]
    state = client.session["oidc_states"][state_key]
    assert state["config_identifier"] == OIDC_ADMIN_CONFIG_IDENTIFIER
    # upstream library
    assert client.session["oidc_login_next"] == "/admin/"
    # our own addition
    assert client.session["oidc-db_redirect_next"] == "/admin/"


@pytest.mark.oidcconfig(oidc_keycloak_idp_hint="keycloak-idp2")
def test_keycloak_idp_hint_via_config(filled_admin_config, settings, client):
    settings.OIDC_KEYCLOAK_IDP_HINT = "keycloak-idp1"
    start_url = reverse("oidc_authentication_init")

    response = client.get(start_url)

    assert response.status_code == 302
    parsed_url = urlsplit(response.url)

    query = parse_qs(parsed_url.query)
    assert query["kc_idp_hint"] == ["keycloak-idp2"]


@pytest.mark.oidcconfig(check_op_availability=True)
def test_check_idp_availability_not_available(
    filled_admin_config, settings, client, requests_mock
):
    requests_mock.get("https://mock-oidc-provider:9999/oidc/auth", status_code=503)
    start_url = reverse("oidc_authentication_init")

    with pytest.raises(OIDCProviderOutage):
        client.get(start_url)


@pytest.mark.oidcconfig(check_op_availability=True)
def test_check_idp_availability_available(
    filled_admin_config, settings, client, requests_mock
):
    requests_mock.get("https://mock-oidc-provider:9999/oidc/auth", status_code=400)
    start_url = reverse("oidc_authentication_init")

    response = client.get(start_url)

    assert response.status_code == 302


@pytest.mark.oidcconfig(oidc_rp_scopes_list=["email"])
@pytest.mark.auth_request
def test_overwrite_scope(dummy_config, auth_request):
    """Test whether the scopes specified in the configuration can be overwritten."""

    @register("test-extra-scope")
    class OIDCTestExtraParamsPlugin(OIDCAdminPlugin):
        def get_extra_params(self, request: HttpRequest, extra_params: dict) -> dict:
            extra_params["scope"] = "not-email and-extra"
            return extra_params

    OIDCClientFactory.create(identifier="test-extra-scope")

    oidc_init = OIDCAuthenticationRequestInitView.as_view(identifier="test-extra-scope")
    redirect_response = oidc_init(auth_request)

    assert redirect_response.status_code == 302

    parsed_url = urlsplit(redirect_response.url)
    query = parse_qs(parsed_url.query)

    assert query["scope"] == ["not-email and-extra"]



"""
Test the init flow.
"""

from urllib.parse import parse_qs, urlsplit

from django.core.exceptions import DisallowedRedirect

import pytest

from mozilla_django_oidc_db.exceptions import OIDCProviderOutage
from mozilla_django_oidc_db.views import OIDCAuthenticationRequestInitView

pytestmark = [pytest.mark.django_db]

init_view = OIDCAuthenticationRequestInitView.as_view(
    identifier="test-oidc", allow_next_from_query=False
)


@pytest.mark.oidcconfig(
    enabled=True,
    oidc_rp_client_id="fixed_client_id",
    oidc_rp_client_secret="supersecret",
    oidc_op_authorization_endpoint="https://example.com/oidc/auth",
)
def test_redirects_to_oidc_provider(dummy_config, auth_request):
    response = init_view(auth_request, return_url="/fixed-next")

    assert response.status_code == 302
    parsed_url = urlsplit(response.url)
    assert parsed_url.scheme == "https"
    assert parsed_url.netloc == "example.com"
    assert parsed_url.path == "/oidc/auth"

    # introspect state
    state_key = parse_qs(parsed_url.query)["state"][0]
    state = auth_request.session["oidc_states"][state_key]
    assert state["config_identifier"] == "test-oidc"
    # upstream library
    assert auth_request.session["oidc_login_next"] == "/fixed-next"
    # our own addition
    assert auth_request.session["oidc-db_redirect_next"] == "/fixed-next"


@pytest.mark.oidcconfig()
def test_suspicious_return_url(dummy_config, auth_request):
    with pytest.raises(DisallowedRedirect):
        init_view(auth_request, return_url="http://evil.com/steal-my-data")


@pytest.mark.parametrize(
    "get_kwargs",
    (
        {},
        {"return_url": ""},
        {"return_url": None},
    ),
)
@pytest.mark.oidcconfig()
def test_forgotten_return_url(dummy_config, auth_request, get_kwargs):
    with pytest.raises(ValueError):
        init_view(auth_request, **get_kwargs)


class IDPCheckInitView(OIDCAuthenticationRequestInitView):
    def check_idp_availability(self) -> None:
        raise OIDCProviderOutage("The internet is bwoken.")


oidc_init_with_idp_check = IDPCheckInitView.as_view(
    identifier="test-oidc", allow_next_from_query=True
)


@pytest.mark.oidcconfig(check_op_availability=True)
def test_idp_check_mechanism(dummy_config, auth_request, settings):
    with pytest.raises(OIDCProviderOutage):
        oidc_init_with_idp_check(auth_request)



from django.contrib import admin
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _

from .forms import OIDCProviderForm
from .models import OIDCClient, OIDCProvider


@admin.register(OIDCClient)
class OIDCClientAdmin(admin.ModelAdmin):
    list_display = (
        "identifier",
        "oidc_rp_client_id",
        "enabled",
    )
    search_fields = ("identifier", "oidc_provider__identifier", "oidc_rp_client_id")
    fieldsets = (
        (
            _("Activation"),
            {"fields": ("enabled",)},
        ),
        (
            _("OIDC Provider"),
            {"fields": ("oidc_provider", "check_op_availability")},
        ),
        (
            _("Relying Party settings"),
            {
                "fields": (
                    "oidc_rp_client_id",
                    "oidc_rp_client_secret",
                    "oidc_rp_scopes_list",
                    "oidc_rp_sign_algo",
                    "oidc_rp_idp_sign_key",
                )
            },
        ),
        (_("Custom settings"), {"fields": ("options",)}),
        (
            _("Advanced settings"),
            {
                "fields": (
                    "userinfo_claims_source",
                    "oidc_keycloak_idp_hint",
                ),
                "classes": [
                    "collapse in",
                ],
            },
        ),
    )

    def has_add_permission(self, request: HttpRequest) -> bool:
        return False

    def has_delete_permission(
        self, request: HttpRequest, obj: OIDCClient | None = None
    ) -> bool:
        return False

    def get_form(self, request, obj=None, change=False, **kwargs):
        form = super().get_form(request, obj, change, **kwargs)

        # The JSON schema for the `options` field needs to be deduced from the instance.
        # django_jsonform field passes the instance to the callable to get the schema
        # if the attribute `instance` is present on the widget instance
        form.base_fields["options"].widget.instance = obj

        return form


@admin.register(OIDCProvider)
class OIDCProviderAdmin(admin.ModelAdmin):
    list_display = ("identifier",)
    search_fields = ("identifier",)
    form = OIDCProviderForm



# Generated by Django 4.2.9 on 2024-05-01 16:10

from django.db import migrations, models

import mozilla_django_oidc_db.fields


class Migration(migrations.Migration):

    dependencies = [
        ("mozilla_django_oidc_db", "0001_initial_to_v015"),
    ]

    operations = [
        migrations.AddField(
            model_name="openidconnectconfig",
            name="new_groups_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                blank=True,
                default=mozilla_django_oidc_db.fields.ClaimFieldDefault("roles"),
                help_text="The name of the OIDC claim that holds the values to map to local user groups.",
                size=None,
                verbose_name="groups claim",
            ),
        ),
        migrations.AddField(
            model_name="openidconnectconfig",
            name="new_username_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                default=mozilla_django_oidc_db.fields.ClaimFieldDefault("sub"),
                help_text="The name of the OIDC claim that is used as the username",
                size=None,
                verbose_name="username claim",
            ),
        ),
        migrations.RemoveField(
            model_name="openidconnectconfig",
            name="groups_claim",
        ),
        migrations.RemoveField(
            model_name="openidconnectconfig",
            name="username_claim",
        ),
        migrations.RenameField(
            model_name="openidconnectconfig",
            old_name="new_groups_claim",
            new_name="groups_claim",
        ),
        migrations.RenameField(
            model_name="openidconnectconfig",
            old_name="new_username_claim",
            new_name="username_claim",
        ),
    ]



from typing import Literal, Union

from django.utils.translation import gettext_lazy as _

from django_setup_configuration.fields import DjangoModelRef
from django_setup_configuration.models import ConfigurationModel
from pydantic import AnyUrl, Discriminator, Field, Tag
from typing_extensions import Annotated

from mozilla_django_oidc_db.models import OIDCClient, OIDCProvider

EXAMPLE_REALM = "http://keycloak.local:8080/realms/test"


class OIDCFullProviderConfig(ConfigurationModel):
    oidc_op_authorization_endpoint: AnyUrl = DjangoModelRef(
        OIDCProvider,
        "oidc_op_authorization_endpoint",
        examples=[f"{EXAMPLE_REALM}/openid-connect/auth"],
    )
    oidc_op_token_endpoint: AnyUrl = DjangoModelRef(
        OIDCProvider,
        "oidc_op_token_endpoint",
        examples=[f"{EXAMPLE_REALM}/protocol/openid-connect/token"],
    )
    oidc_op_user_endpoint: AnyUrl = DjangoModelRef(
        OIDCProvider,
        "oidc_op_user_endpoint",
        examples=[f"{EXAMPLE_REALM}/protocol/openid-connect/userinfo"],
    )
    oidc_op_logout_endpoint: AnyUrl | Literal[""] = DjangoModelRef(
        OIDCProvider,
        "oidc_op_logout_endpoint",
        examples=[f"{EXAMPLE_REALM}/protocol/openid-connect/logout"],
    )
    oidc_op_jwks_endpoint: AnyUrl | Literal[""] = DjangoModelRef(
        OIDCProvider,
        "oidc_op_jwks_endpoint",
        examples=[f"{EXAMPLE_REALM}/protocol/openid-connect/certs"],
    )


class OIDCDiscoveryProviderConfig(ConfigurationModel):
    oidc_op_discovery_endpoint: AnyUrl = DjangoModelRef(
        OIDCProvider,
        "oidc_op_discovery_endpoint",
        examples=[f"{EXAMPLE_REALM}/"],
    )


def get_provider_config_model(endpoint_data):
    """Get which configuration model to use for the OIDC provider"""
    if isinstance(endpoint_data, dict):
        discovery_endpoint = endpoint_data.get("oidc_op_discovery_endpoint")
    else:
        discovery_endpoint = getattr(endpoint_data, "oidc_op_discovery_endpoint", None)
    if discovery_endpoint:
        return "discovery"
    return "all"


OIDCProviderConfigUnion = Annotated[
    Union[
        Annotated[OIDCFullProviderConfig, Tag("all")],
        Annotated[OIDCDiscoveryProviderConfig, Tag("discovery")],
    ],
    Discriminator(get_provider_config_model),
]


class OIDCConfigProviderModel(ConfigurationModel):
    identifier: str = DjangoModelRef(
        OIDCProvider,
        "identifier",
        description="a unique identifier for this OIDC provider.",
        examples=["admin-oidc-provider"],
    )
    endpoint_config: OIDCProviderConfigUnion

    class Meta:
        django_model_refs = {
            OIDCProvider: [
                "oidc_token_use_basic_auth",
                "oidc_use_nonce",
                "oidc_nonce_size",
                "oidc_state_size",
            ]
        }


class AdminOIDCConfigurationModelItem(ConfigurationModel):
    identifier: str = DjangoModelRef(
        OIDCClient,
        "identifier",
        description="a unique identifier for this configuration",
        examples=["admin-oidc"],
    )

    enabled: bool = DjangoModelRef(OIDCClient, "enabled", default=True)
    oidc_rp_scopes_list: list[str] = DjangoModelRef(OIDCClient, "oidc_rp_scopes_list")
    options: dict = DjangoModelRef(OIDCClient, "options", default_factory=dict)

    endpoint_config: OIDCProviderConfigUnion | None = Field(
        description=_("Configuration for the OIDC Provider endpoints."),
        default=None,
        deprecated=True,
    )
    oidc_provider_identifier: str = DjangoModelRef(
        OIDCProvider, "identifier", examples=["test-oidc-provider"], default=""
    )

    ## DEPRECATED FIELDS
    claim_mapping: dict = Field(
        default_factory=lambda: {
            "email": ["email"],
            "first_name": ["given_name"],
            "last_name": ["family_name"],
        },
        description=_("Mapping from User model field names to a path in the claim."),
        deprecated=True,
    )
    oidc_token_use_basic_auth: bool = Field(
        default=False,
        description=_(
            "If enabled, the client ID and secret are sent in the HTTP Basic auth "
            "header when obtaining the access token. Otherwise, they are sent in the "
            "request body."
        ),
        deprecated=True,
    )
    oidc_use_nonce: bool = Field(
        default=True,
        description=_("Controls whether the client uses nonce verification"),
        deprecated=True,
    )
    oidc_nonce_size: int = Field(
        default=32,
        description=_(
            "Sets the length of the random string used for nonce verification"
        ),
        deprecated=True,
    )
    oidc_state_size: int = Field(
        default=32,
        description=_(
            "Sets the length of the random string used for state verification"
        ),
        deprecated=True,
    )
    # Arrays are overridden to make the typing simpler (the underlying Django field is an ArrayField, which is non-standard)
    username_claim: list[str] = Field(
        default_factory=lambda: ["sub"],
        description=_("Path in the claims to the value to use as username."),
        deprecated=True,
        examples=[["nested", "username", "claim"]],
    )
    groups_claim: list[str] = Field(
        default_factory=lambda: ["roles"],
        description=_("Path in the claims to the value with group names."),
        deprecated=True,
        examples=[["nested", "group", "claim"]],
    )
    superuser_group_names: list[str] = Field(
        default_factory=list,
        description=_("Superuser group names"),
        deprecated=True,
        examples=[["superusers"]],
    )
    default_groups: list[str] = Field(
        default_factory=list,
        description=_("Default group names"),
        deprecated=True,
        examples=[["read-only-users"]],
    )
    sync_groups: bool = Field(
        description=_("Whether to sync local groups"),
        deprecated=True,
        examples=[True],
        default=True,
    )
    sync_groups_glob_pattern: str = Field(
        description=_("Pattern that the group names to sync should follow."),
        deprecated=True,
        examples=["*"],
        default="*",
    )
    make_users_staff: bool = Field(
        description=_("Whether to make the users staff."),
        deprecated=True,
        examples=[False],
        default=False,
    )

    class Meta:
        django_model_refs = {
            OIDCClient: [
                "oidc_rp_client_id",
                "oidc_rp_client_secret",
                "oidc_rp_sign_algo",
                "oidc_rp_idp_sign_key",
                "oidc_keycloak_idp_hint",
                "userinfo_claims_source",
            ]
        }
        extra_kwargs = {
            "oidc_rp_client_id": {"examples": ["modify-this"]},
            "oidc_rp_client_secret": {"examples": ["modify-this"]},
            "oidc_rp_idp_sign_key": {"examples": ["modify-this"]},
            "oidc_keycloak_idp_hint": {"examples": ["some-identity-provider"]},
        }


class AdminOIDCConfigurationModel(ConfigurationModel):
    providers: list[OIDCConfigProviderModel] = Field(
        default_factory=list, description=_("List of OIDC providers")
    )
    items: list[AdminOIDCConfigurationModelItem]



import logging

from django.contrib.auth.models import Group, User
from django.core.exceptions import BadRequest, ImproperlyConfigured
from django.http import HttpRequest
from django.test import RequestFactory

import pytest
from mozilla_django_oidc.auth import OIDCAuthenticationBackend as UpstreamBackend

from mozilla_django_oidc_db.backends import OIDCAuthenticationBackend
from mozilla_django_oidc_db.models import (
    OIDCClient,
    OIDCProvider,
    UserInformationClaimsSources,
)
from mozilla_django_oidc_db.views import OIDCAuthenticationRequestInitView
from testapp.backends import MockBackend

#
# DYNAMIC CONFIGURATION TESTS
#


@pytest.mark.oidcconfig(enabled=False)
def test_authenticate_oidc_not_enabled(dummy_config, callback_request: HttpRequest):
    backend = OIDCAuthenticationBackend()

    user = backend.authenticate(callback_request)

    # Authentication with the backend should not return a result,
    # because OIDC is not enabled
    assert user is None


@pytest.mark.django_db
@pytest.mark.oidcconfig
@pytest.mark.callback_request(
    init_view=OIDCAuthenticationRequestInitView.as_view(identifier="test-oidc-disabled")
)
def test_authentication_loads_config_from_init_state(
    dummy_config, disabled_config, callback_request: HttpRequest
):
    assert dummy_config.enabled is True
    backend = OIDCAuthenticationBackend()
    # check that we correctly initialized our OIDC state
    state_key = callback_request.GET["state"]
    state = callback_request.session["oidc_states"][state_key]
    assert state["config_identifier"] == "test-oidc-disabled"

    user = backend.authenticate(callback_request)

    # Authentication with the backend should not return a result,
    # because OIDC is not enabled
    assert user is None


@pytest.mark.parametrize("sign_alg", ("RS**", "ES**"))
@pytest.mark.django_db
def test_settings_still_validated(settings, sign_alg: str):
    """
    Test that the upstream library settings checks are still performed.
    """
    config = OIDCClient.objects.get(identifier="test-oidc-not-configured")
    oidc_provider = OIDCProvider.objects.create(
        identifier="test-not-configured-provider", oidc_op_jwks_endpoint=""
    )
    config.oidc_rp_sign_algo = sign_alg
    config.oidc_rp_idp_sign_key = ""
    config.oidc_provider = oidc_provider
    config.save()
    backend = OIDCAuthenticationBackend()
    backend.config = config

    with pytest.raises(ImproperlyConfigured):
        setting = backend.OIDC_RP_CLIENT_ID  # the exact setting doesn't matter

    # check that the same error is raised by the upstream backend
    settings.OIDC_RP_SIGN_ALGO = sign_alg
    settings.OIDC_RP_IDP_SIGN_KEY = None
    settings.OIDC_OP_JWKS_ENDPOINT = None
    with pytest.raises(ImproperlyConfigured):
        UpstreamBackend()


#
# LOGGING AND OBFUSCATION TESTS
#


@pytest.mark.oidcconfig(
    enabled=True,
    extra_options={
        "user_settings.claim_mappings.username": ["sub"],
        "user_settings.sensitive_claims": [
            ["sensitive_claim1"],
            ["parent", "sensitive_claim2"],
        ],
    },
)
def test_obfuscates_sensitive_claims(dummy_config, caplog):
    caplog.set_level(logging.DEBUG, logger="mozilla_django_oidc_db.plugins")
    backend = OIDCAuthenticationBackend()
    backend.config = dummy_config

    claims_ok = backend.verify_claims(
        {
            "sub": "some-unique-id",
            "sensitive_claim1": "obfuscate-me!",
            "parent": {
                "sensitive_claim2": "obfuscate-me!",
                "other": "not-sensitive",
            },
        }
    )

    assert claims_ok

    # check that the emitted log records have the sensitive claims obfuscated
    log_message: str = next(
        message
        for rec in caplog.records
        if (message := rec.message).startswith("OIDC claims received")
    )

    assert "some-unique-id" not in log_message
    assert "obfuscate-me!" not in log_message
    assert "not-sensitive" in log_message


#
# USER CREATION & UPDATING TESTS
#


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    extra_options={
        "user_settings.claim_mappings.first_name": ["given_name"],
        "user_settings.claim_mappings.last_name": ["family_name"],
    },
)
def test_create_user_with_default_config(dummy_config, callback_request: HttpRequest):
    assert not User.objects.exists()
    backend = MockBackend(
        claims={
            "sub": "123456",
            "email": "admin@localhost",
            "given_name": "John",
            "family_name": "Doe",
        }
    )

    user = backend.authenticate(request=callback_request)

    assert user is not None
    assert isinstance(user, User)

    # Verify that a user is created with the correct values
    assert user.pk is not None
    assert user.username == "123456"
    assert user.email == "admin@localhost"
    assert user.first_name == "John"
    assert user.last_name == "Doe"


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    extra_options={
        "user_settings.claim_mappings.first_name": ["given_name"],
        "user_settings.claim_mappings.last_name": ["family_name"],
        "user_settings.username_case_sensitive": False,
    },
)
def test_case_insensitive_username_lookups(
    settings, dummy_config, callback_request: HttpRequest
):
    existing_user = User.objects.create(username="ADMIN")

    backend = MockBackend(
        claims={
            "sub": "admin",
            "email": "admin@localhost",
            "given_name": "John",
            "family_name": "Doe",
        }
    )

    user = backend.authenticate(request=callback_request)

    assert user == existing_user
    assert User.objects.count() == 1


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    extra_options={
        "user_settings.claim_mappings.username": ["user", "username"],
        "user_settings.claim_mappings.first_name": ["profile", "given_name"],
        "user_settings.claim_mappings.last_name": ["profile", "family_name"],
        "user_settings.claim_mappings.email": ["attributes.email"],
        "groups_settings.claim_mapping": ["user", "groups.names"],
    },
)
def test_create_user_with_custom_and_complex_config(
    dummy_config, callback_request: HttpRequest
):
    assert not User.objects.exists()
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "user": {
                "username": "admin",
                "groups.names": ["supergroup"],
            },
            "profile": {
                "given_name": "John",
                "family_name": "Doe",
            },
            "attributes.email": "admin@localhost",
        }
    )

    user = backend.authenticate(request=callback_request)

    assert user is not None
    assert isinstance(user, User)

    # Verify that a user is created with the correct values
    assert user.pk is not None
    assert user.username == "admin"
    assert user.email == "admin@localhost"
    assert user.first_name == "John"
    assert user.last_name == "Doe"


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    extra_options={
        "user_settings.claim_mappings.username": ["user", "username"],
        "groups_settings.claim_mapping": ["user", "groups.names"],
        "groups_settings.sync": True,
    },
)
def test_create_user_with_mapped_groups(dummy_config, callback_request: HttpRequest):
    assert not User.objects.exists()
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "user": {
                "username": "admin",
                "groups.names": ["supergroup"],
            },
            "email": "admin@localhost",
            "given_name": "John",
            "family_name": "Doe",
        }
    )

    user = backend.authenticate(request=callback_request)

    assert user is not None
    assert isinstance(user, User)

    group_names = set(user.groups.values_list("name", flat=True))
    assert group_names == {"supergroup"}


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    extra_options={
        "user_settings.claim_mappings.username": ["sub"],
        "groups_settings.claim_mapping": ["roles"],
        "groups_settings.sync": True,
    },
)
def test_groups_claim_string_instead_of_list(
    dummy_config, callback_request: HttpRequest
):
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "sub": "admin",
            "email": "admin@localhost",
            "given_name": "John",
            "family_name": "Doe",
            "roles": "admins",
        }
    )

    user = backend.authenticate(request=callback_request)

    assert user is not None
    assert isinstance(user, User)

    assert set(user.groups.values_list("name", flat=True)) == {"admins"}


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    extra_options={
        "user_settings.claim_mappings.username": ["user", "username"],
        "user_settings.claim_mappings.email": ["attributes.email"],
        "user_settings.claim_mappings.first_name": ["profile", "given_name"],
        "user_settings.claim_mappings.last_name": ["profile", "family_name"],
        "groups_settings.claim_mapping": ["user", "groups.names"],
        "groups_settings.sync": True,
        "groups_settings.sync_pattern": "*",
        "groups_settings.default_groups": [],  # explicitly clear to wipe any possible implicit defaults
    },
)
def test_update_user_with_custom_and_complex_config(
    dummy_config, callback_request: HttpRequest
):
    existing_user = User.objects.create(
        username="admin",
        email="outdated@example.com",
        first_name="O.",
        last_name="Utdated",
        is_superuser=True,
    )
    group = Group.objects.create(name="to-remove")
    existing_user.groups.add(group)

    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "user": {
                "username": "admin",
                "groups.names": ["supergroup"],
            },
            "profile": {
                "given_name": "John",
                "family_name": "Doe",
            },
            "attributes.email": "admin@localhost",
        }
    )

    user = backend.authenticate(request=callback_request)

    assert isinstance(user, User)

    # check that the existing user was correctly found based on the username claim
    assert user == existing_user

    # Verify that a user is created with the correct values
    assert user.username == "admin"
    assert user.email == "admin@localhost"
    assert user.first_name == "John"
    assert user.last_name == "Doe"
    # is_superuser may not be affected if `superuser_group_names` is not set
    assert user.is_superuser

    # missing groups must be created because of sync_groups and sync_groups_glob_pattern
    assert set(user.groups.values_list("name", flat=True)) == {"supergroup"}


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    extra_options={
        "groups_settings.claim_mapping": [],
        "groups_settings.sync": True,
        "groups_settings.sync_pattern": "*",
    },
)
def test_authenticate_user_no_group_sync_without_claim(
    dummy_config, callback_request: HttpRequest
):
    Group.objects.create(name="group1")
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "sub": "123456",
            "roles": ["group1", "newgroup"],
        }
    )

    user = backend.authenticate(request=callback_request)

    assert isinstance(user, User)
    # Verify that no groups were created or assigned
    assert Group.objects.count() == 1
    assert not user.groups.exists()


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    extra_options={
        "groups_settings.claim_mapping": ["roles"],
        "groups_settings.sync": True,
        "groups_settings.sync_pattern": "myapp:*",
    },
)
def test_authenticate_user_groups_glob_pattern(
    dummy_config, callback_request: HttpRequest
):
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "sub": "123456",
            "roles": ["admin", "myapp:editor"],
        }
    )

    user = backend.authenticate(request=callback_request)

    assert isinstance(user, User)
    # Verify that no groups were created or assigned
    assert Group.objects.count() == 1
    assert set(user.groups.values_list("name", flat=True)) == {"myapp:editor"}


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    extra_options={
        "groups_settings.claim_mapping": ["roles"],
        "groups_settings.sync": True,
        "groups_settings.sync_pattern": "myapp:*",
        "groups_settings.default_groups": ["default1"],
    },
)
def test_authenticate_user_groups_and_default_groups(
    dummy_config, callback_request: HttpRequest
):
    Group.objects.create(name="default1")
    Group.objects.create(name="default2")
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "sub": "123456",
            "roles": ["admin", "myapp:editor"],
        }
    )

    user = backend.authenticate(request=callback_request)

    assert isinstance(user, User)
    # Verify that no groups were created or assigned
    assert Group.objects.count() == 3
    assert set(user.groups.values_list("name", flat=True)) == {
        "myapp:editor",
        "default1",
    }


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    extra_options={
        "user_settings.claim_mappings.is_superuser": ["is_god"],
        "groups_settings.make_users_staff": True,
    },
)
def test_authenticate_user_make_staff(dummy_config, callback_request: HttpRequest):
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "sub": "123456",
            "is_god": 1,
        }
    )

    user = backend.authenticate(request=callback_request)

    assert isinstance(user, User)
    assert user.is_staff
    assert user.is_superuser


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    extra_options={
        "groups_settings.make_users_staff": True,
        "groups_settings.claim_mapping": ["roles"],
        "groups_settings.sync": False,
        "groups_settings.superuser_group_names": ["superuser"],
    },
)
def test_authenticate_user_make_superuser_based_on_group(
    dummy_config, callback_request: HttpRequest
):
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "sub": "123456",
            "roles": ["superuser", "groupadmin"],
        }
    )

    user = backend.authenticate(request=callback_request)

    assert isinstance(user, User)
    assert user.is_superuser
    assert not Group.objects.exists()


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    extra_options={
        "groups_settings.make_users_staff": True,
        "groups_settings.claim_mapping": ["roles"],
        "groups_settings.sync": False,
        "groups_settings.superuser_group_names": ["superuser"],
    },
)
def test_remove_superuser_based_on_group(dummy_config, callback_request: HttpRequest):
    existing_user = User.objects.create_user(username="123456", is_superuser=True)
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "sub": "123456",
            "roles": ["nosuperuser", "groupadmin"],
        }
    )

    user = backend.authenticate(request=callback_request)

    assert isinstance(user, User)
    assert user == existing_user
    assert not user.is_superuser


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    extra_options={
        "groups_settings.make_users_staff": True,
        "groups_settings.claim_mapping": ["roles"],
        "groups_settings.sync": False,
        "groups_settings.superuser_group_names": [],
    },
)
def test_do_nothing_if_no_superuser_groups_configured(
    dummy_config, callback_request: HttpRequest
):
    existing_user = User.objects.create_user(username="123456", is_superuser=True)
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "sub": "123456",
            "roles": ["nosuperuser", "groupadmin"],
        }
    )

    user = backend.authenticate(request=callback_request)

    assert isinstance(user, User)
    assert user == existing_user
    assert user.is_superuser


#
# AUTHENTICATE FLOW PROBLEMS
#


@pytest.mark.oidcconfig(enabled=True)
def test_authenticate_called_without_args(dummy_config):
    backend = OIDCAuthenticationBackend()

    user = backend.authenticate(request=None)

    assert user is None


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    extra_options={
        "user_settings.claim_mappings.username": ["sub"],
    },
)
def test_username_claim_empty(dummy_config, callback_request: HttpRequest):
    backend = MockBackend(claims={"sub": ""})

    user = backend.authenticate(request=callback_request)

    assert user is None


@pytest.mark.oidcconfig(
    enabled=True, userinfo_claims_source=UserInformationClaimsSources.id_token
)
def test_empty_claims(dummy_config, callback_request: HttpRequest):
    backend = MockBackend(claims={})

    user = backend.authenticate(request=callback_request)

    assert user is None


@pytest.mark.oidcconfig(
    enabled=True,
    userinfo_claims_source=UserInformationClaimsSources.id_token,
    extra_options={
        "user_settings.claim_mappings.username": ["sub"],
        "groups_settings.claim_mapping": ["roles"],
        "groups_settings.sync": True,
    },
)
def test_groups_claim_wrong_type(dummy_config, callback_request: HttpRequest):
    backend = MockBackend(
        # a variation of nested paths, custom claims and dots in claim names
        claims={
            "sub": "admin",
            "email": "admin@localhost",
            "given_name": "John",
            "family_name": "Doe",
            "roles": ["group1", None, 123],
        }
    )

    user = backend.authenticate(request=callback_request)

    assert user is not None
    assert isinstance(user, User)

    # fails silently
    assert not user.groups.exists()


def test_authenticate_without_previous_state(rf: RequestFactory):
    request = rf.get("/oidc/callback", {"state": "foo", "code": "bar"})
    request.session = {}
    backend = OIDCAuthenticationBackend()

    with pytest.raises(BadRequest):
        backend.authenticate(request=request)


#
# CACHE/QUERYING FOR CONFIG DURING PERMISSON CHECK.
#


def test_init_does_not_perform_config_io(mocker):
    """
    Regression test for https://github.com/maykinmedia/mozilla-django-oidc-db/issues/30

    This test will fail if IO is performed in one of two ways:

    * assertion fails because get_solo was called
    * pytest will complain about database access which is forbidden because there is
      no pytest.mark.django_db present (deliberately)
    """
    m_get_setting = mocker.patch(
        "mozilla_django_oidc_db.config.get_setting_from_config"
    )

    # instantiate
    OIDCAuthenticationBackend()

    m_get_setting.assert_not_called()



from django.utils.translation import gettext as _

from mozilla_django_oidc_db.fields import ClaimFieldDefault


def test_claim_field_default_equality():
    assert ClaimFieldDefault("foo", "bar") == ClaimFieldDefault("foo", "bar")
    assert ClaimFieldDefault("foo", "bar") != ClaimFieldDefault("bar", "foo")
